import bisect
import dataclasses

from libresvip.core.constants import DEFAULT_PHONEME
from libresvip.core.lyric_phoneme.chinese import CHINESE_RE, get_pinyin_series
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info

from .model import (
    AISAudioPattern,
    AISAudioTrack,
    AISNote,
    AISProjectBody,
    AISProjectHead,
    AISSingVoicePattern,
    AISSingVoiceTrack,
    AISTempo,
    AISTimeSignature,
)
from .options import OutputOptions


@dataclasses.dataclass
class AiSingersGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> tuple[AISProjectHead, AISProjectBody]:
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        ais_time_signatures = self.generate_time_signatures(project.time_signature_list)
        ais_tempos = self.generate_tempos(project.song_tempo_list, project.time_signature_list)
        max_end_time = round(
            max(
                (
                    note.end_pos
                    for track in project.track_list
                    if isinstance(track, SingingTrack)
                    for note in track.note_list
                ),
                default=project.song_tempo_list[-1].position,
            )
            / 15
        )
        num_bars = ais_tempos[-1].start_bar + (max_end_time - ais_tempos[-1].start_128) // (
            128 * ais_time_signatures[-1].beat_zi / ais_time_signatures[-1].beat_mu
        )
        ais_project_head = AISProjectHead(
            signature=ais_time_signatures,
            tempo=ais_tempos,
            time=max_end_time,
            bar=max(num_bars, 100),
        )
        ais_project_body = AISProjectBody(tracks=self.generate_tracks(project.track_list))
        return ais_project_head, ais_project_body

    def generate_time_signatures(
        self, time_signatures: list[TimeSignature]
    ) -> list[AISTimeSignature]:
        return [
            AISTimeSignature(
                beat_zi=time_signature.numerator,
                beat_mu=time_signature.denominator,
                start_bar=time_signature.bar_index,
            )
            for time_signature in time_signatures
        ]

    def generate_tempos(
        self, tempos: list[SongTempo], time_signatures: list[TimeSignature]
    ) -> list[AISTempo]:
        ais_tempos = []
        prev_bar_index = 1
        cur_tick = 0
        tick_indexes = []
        for i, time_signature in enumerate(time_signatures):
            if time_signature.bar_index > prev_bar_index and i:
                cur_tick += time_signature[i - 1].bar_length() * (
                    time_signature.bar_index - prev_bar_index
                )
            tick_indexes.append(cur_tick)
            if time_signature.bar_index > prev_bar_index:
                prev_bar_index = time_signature.bar_index
        for tempo in tempos:
            ts_index = min(
                bisect.bisect_left(tick_indexes, tempo.position),
                len(tick_indexes) - 1,
            )
            start_bar = max(
                (
                    time_signatures[ts_index].bar_index
                    + (tempo.position - tick_indexes[ts_index] - self.first_bar_length)
                    // time_signatures[ts_index].bar_length()
                ),
                0,
            )
            start_beat_in_bar = (
                (tempo.position - tick_indexes[ts_index] - self.first_bar_length)
                % time_signatures[ts_index].bar_length()
            ) // (time_signatures[ts_index].bar_length() / time_signatures[ts_index].numerator)
            ais_tempo = AISTempo(
                tempo_float=tempo.bpm,
                start_128=round(tempo.position / 15),
                start_bar=start_bar,
                start_beat_in_bar=start_beat_in_bar,
            )
            ais_tempos.append(ais_tempo)
        return ais_tempos

    def generate_tracks(self, tracks: list[Track]) -> list[AISSingVoiceTrack]:
        ais_tracks: list[AISSingVoiceTrack] = []
        for track in tracks:
            if isinstance(track, SingingTrack):
                if note_list := self.generate_notes(track):
                    ais_track = AISSingVoiceTrack(
                        idx=len(ais_tracks),
                        name=track.title,
                        mute=track.mute,
                        solo=track.solo,
                        singer_namecn=track.ai_singer_name,
                        items=[
                            AISSingVoicePattern(
                                uid=len(tracks) + len(ais_tracks),
                                start=0,
                                length=max(note.start + note.length for note in note_list)
                                + self.first_bar_length,
                                notes=note_list,
                            )
                        ],
                    )
                    ais_tracks.append(ais_track)
            elif isinstance(track, InstrumentalTrack) and (
                track_info := audio_track_info(track.audio_file_path, only_wav=True)
            ):
                offset_secs = track_info.duration / 1000
                end_tick = self.synchronizer.get_actual_ticks_from_secs_offset(
                    track.offset, offset_secs
                )
                ais_track = AISAudioTrack(
                    idx=len(ais_tracks),
                    name=track.title,
                    mute=track.mute,
                    solo=track.solo,
                    items=[
                        AISAudioPattern(
                            start=track.offset // 15,
                            length=(end_tick - track.offset) // 15,
                            path_audio=track.audio_file_path,
                            path_wave=track.audio_file_path,
                            len_sec=int(offset_secs),
                            n_channel=track_info.channel_s,
                        )
                    ],
                )
                ais_tracks.append(ais_track)
        return ais_tracks

    def generate_notes(self, track: SingingTrack) -> list[AISNote]:
        ais_notes = []
        for note in track.note_list:
            note_start = int(note.start_pos / 15)
            ais_note = AISNote(
                midi_no=note.key_number - 12,
                start=note_start,
                length=int(note.end_pos / 15) - note_start,
                lyric=note.lyric,
                pinyin=note.pronunciation
                or (
                    " ".join(get_pinyin_series(note.lyric))
                    if CHINESE_RE.fullmatch(note.lyric) is not None
                    else (note.lyric or DEFAULT_PHONEME)
                ),
                triple=False,
                pit="0x500",
            )
            if pitch_points := self.generate_pitch(track.edited_params.pitch, note):
                ais_note.pit = pitch_points
            ais_notes.append(ais_note)
        return ais_notes

    def generate_pitch(self, pitch_param_curve: ParamCurve, note: Note) -> list[float]:
        tick_step = note.length / 500.0
        sample_time_list = [
            note.start_pos + self.first_bar_length + int(tick_step * i) for i in range(500)
        ]
        pitch_param_in_note = [
            p
            for p in pitch_param_curve.points.root
            if note.start_pos + self.first_bar_length <= p.x < note.end_pos + self.first_bar_length
        ]

        pitch_param_time_in_note = dict(pitch_param_in_note)

        ais_pitch_param: list[int | float] = []
        for sample_time in sample_time_list:
            if (pitch := pitch_param_time_in_note.get(sample_time)) is None:
                distance = -1
                value = 0.0

                for point in pitch_param_in_note:
                    if distance > abs(point.x - sample_time) or distance == -1:
                        distance = abs(point.x - sample_time)
                        value = 0 if point.y == -100 else (point.y - note.key_number * 100) / 10
                ais_pitch_param.append(value)

            elif pitch == -100:
                ais_pitch_param.append(0)
            else:
                ais_pitch_param.append((pitch - note.key_number * 100) / 10)
        buffer = []
        previous_node = ais_pitch_param[0]
        previous_node_index = 0
        for i in range(len(ais_pitch_param)):
            if ais_pitch_param[i] == previous_node:
                buffer.append(ais_pitch_param[i])
            else:
                for j in range(len(buffer)):
                    ais_pitch_param[previous_node_index + j] = previous_node + j * (
                        ais_pitch_param[i] - buffer[j]
                    ) / len(buffer)
                buffer.clear()

            if ais_pitch_param[i] != previous_node:
                previous_node_index = i
                previous_node = ais_pitch_param[i]

        return ais_pitch_param



import abc

from pydantic import BaseModel, Field

from libresvip.core.constants import DEFAULT_BPM
from libresvip.utils.translation import gettext_lazy as _


class EnableInstrumentalTrackImportationMixin(BaseModel, abc.ABC):
    import_instrumental_track: bool = Field(True, title=_("Import instrumental tracks"))


class EnablePitchImportationMixin(BaseModel, abc.ABC):
    import_pitch: bool = Field(True, title=_("Import pitch curve"))


class EnableVolumeImportationMixin(BaseModel, abc.ABC):
    import_volume: bool = Field(True, title=_("Import dynamics envelope"))


class EnableBreathImportationMixin(BaseModel, abc.ABC):
    import_breath: bool = Field(True, title=_("Import breath envelope"))


class EnableGenderImportationMixin(BaseModel, abc.ABC):
    import_gender: bool = Field(True, title=_("Import gender envelope"))


class EnableStrengthImportationMixin(BaseModel, abc.ABC):
    import_strength: bool = Field(True, title=_("Import strength envelope"))


class StaticTempoMixin(BaseModel, abc.ABC):
    tempo: float = Field(
        default=DEFAULT_BPM,
        title=_("Constant tempo"),
        description=_(
            "Use this tempo to reset time axis of projects with dynamic tempos or without tempos"
        ),
    )


class SelectSingleTrackMixin(BaseModel, abc.ABC):
    track_index: int = Field(
        default=-1,
        title=_("Track index"),
        description=_("Start from 0, -1 means automatic selection"),
    )


class ExtractEmbededAudioMixin(BaseModel, abc.ABC):
    extract_audio: bool = Field(
        False,
        title=_("Extract audio"),
        description=_(
            "Extract audio from project file for later use if the path does not exist (this option has the side effect of writing additional files)"
        ),
    )



from __future__ import annotations

import dataclasses
import operator
from functools import reduce, singledispatchmethod
from typing import TYPE_CHECKING

import portion

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping

    UnaryFunction = Callable[[int | float], float]
    UnaryFunctionOrConstant = UnaryFunction | int | float


class PiecewiseIntervalDict(portion.IntervalDict):
    def __init__(
        self,
        mapping_or_iterable: Mapping[portion.Interval, UnaryFunctionOrConstant]
        | Iterable[tuple[portion.Interval, UnaryFunctionOrConstant]]
        | None = None,
    ) -> None:
        super().__init__(mapping_or_iterable=mapping_or_iterable)
        self._last_index = 0

    def _get_func(self, x: float) -> UnaryFunctionOrConstant | None:
        _last_index = self._last_index
        while _last_index < len(self._storage._list):
            boundary, func = self._storage.peekitem(_last_index)
            if x in boundary:
                self._last_index = _last_index
                return func
            elif x < boundary:
                self._last_index = _last_index
                break
            _last_index += 1

    def __setitem__(self, key: portion.Interval | float, value: UnaryFunctionOrConstant) -> None:
        if isinstance(key, portion.Interval):
            interval = key
        else:
            interval = self._klass.from_atomic(portion.Bound.CLOSED, key, key, portion.Bound.CLOSED)

        if interval.empty:
            return

        self._storage[interval] = value

    def __getitem__(self, key: portion.Interval | float) -> UnaryFunctionOrConstant | None:
        if isinstance(key, portion.Interval):
            return super().__getitem__(key)
        elif (func := self._get_func(key)) is not None:
            return func(key) if callable(func) else func
        raise KeyError(key)


@dataclasses.dataclass
class RangeInterval:
    _sub_ranges: dataclasses.InitVar[list[tuple[int, int]] | None] = None
    interval: portion.Interval = dataclasses.field(init=False)

    def __post_init__(self, _sub_ranges: list[tuple[int, int]] | None) -> None:
        self.interval = reduce(
            operator.or_,
            (portion.closedopen(*sub_range) for sub_range in (_sub_ranges or [])),
            portion.empty(),
        )

    def is_empty(self) -> bool:
        return self.interval.empty

    def sub_ranges(self) -> Iterable[tuple[int, int]]:
        for sub_range in portion.to_data(self.interval):
            yield sub_range[1], sub_range[2]

    def sub_range_including(self, value: int) -> tuple[int, int] | None:
        for sub_range in self.interval:
            if value in sub_range:
                sub_range_data = portion.to_data(sub_range)
                return sub_range_data[1], sub_range_data[2]
        return None

    def includes(self, value: int) -> bool:
        return value in self.interval

    def __contains__(self, interval: RangeInterval | int) -> bool:
        if isinstance(interval, RangeInterval):
            interval = interval.interval
        return interval in self.interval

    def intersection(self, interval: RangeInterval) -> RangeInterval:
        new_interval = RangeInterval()
        new_interval.interval = self.interval & interval.interval
        return new_interval

    def union(self, interval: RangeInterval) -> RangeInterval:
        new_interval = RangeInterval()
        new_interval.interval = self.interval | interval.interval
        return new_interval

    def sub(self, interval: RangeInterval) -> RangeInterval:
        new_interval = RangeInterval()
        new_interval.interval = self.interval - interval.interval
        return new_interval

    def complement(self, complete_interval: RangeInterval) -> RangeInterval:
        new_interval = RangeInterval()
        new_interval.interval = complete_interval.interval - self.interval
        return new_interval

    @singledispatchmethod
    def expand(self, a: tuple[int, int] | int) -> RangeInterval:
        raise NotImplementedError

    @expand.register(tuple)
    def _(self, radius_tuple: tuple[int, int]) -> RangeInterval:
        return RangeInterval(
            [
                (
                    sub_range[0] - radius_tuple[0],
                    sub_range[1] + radius_tuple[1],
                )
                for sub_range in self.sub_ranges()
            ]
        )

    @expand.register(int)
    def _(self, radius: int) -> RangeInterval:
        return self.expand((radius, radius))

    @singledispatchmethod
    def shrink(self, a: tuple[int, int] | int) -> RangeInterval:
        raise NotImplementedError

    @shrink.register(int)
    def _(self, radius: int) -> RangeInterval:
        return self.expand((-radius, -radius))

    @shrink.register(tuple)
    def _(self, radius_tuple: tuple[int, int]) -> RangeInterval:
        return self.expand((-radius_tuple[0], -radius_tuple[1]))

    def shift(self, offset: int) -> RangeInterval:
        return RangeInterval(
            [(sub_range[0] + offset, sub_range[1] + offset) for sub_range in self.sub_ranges()]
        )

    def __and__(self, other: RangeInterval) -> RangeInterval:
        return self.intersection(other)

    def __or__(self, other: RangeInterval) -> RangeInterval:
        return self.union(other)

    def __sub__(self, other: RangeInterval) -> RangeInterval:
        return self.sub(other)

    def __xor__(self, other: RangeInterval) -> RangeInterval:
        return (self | other) - (self & other)

    def __rshift__(self, other: int) -> RangeInterval:
        return self.shift(other)

    def __lshift__(self, other: int) -> RangeInterval:
        return self.shift(-other)



import io
import pathlib

from libresvip.core.compat import ZipFile, json
from libresvip.core.exceptions import InvalidFileTypeError
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .legacy_model import PpsfLegacyProject
from .model import PpsfProject
from .options import InputOptions, OutputOptions
from .piapro_studio_generator import PiaproStudioGenerator
from .piapro_studio_legacy_parser import PiaproStudioLegacyParser
from .piapro_studio_nt_parser import PiaproStudioNTParser


class PiaproStudioConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        content = path.read_bytes()
        if content[:2] == b"PK":
            with ZipFile(io.BytesIO(content), "r") as zf:
                proj_text = zf.read("ppsf.json")
            ppsf_project = PpsfProject.model_validate_json(proj_text)
            return PiaproStudioNTParser(options).parse_project(ppsf_project)
        elif content[:4] == b"PPSF":
            ppsf_project = PpsfLegacyProject.parse(content)
            return PiaproStudioLegacyParser(options).parse_project(ppsf_project)
        else:
            msg = "Unrecognizable format"
            raise InvalidFileTypeError(msg)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        ppsf_project = PiaproStudioGenerator(options).generate_project(project)
        proj_text = json.dumps(
            ppsf_project.model_dump(
                by_alias=True,
                exclude_none=True,
                mode="json",
            ),
            ensure_ascii=False,
            separators=(", ", ": "),
        )
        buffer = io.BytesIO()
        with ZipFile(buffer, "w") as zf:
            zf.writestr("ppsf.json", proj_text)
        path.write_bytes(buffer.getvalue())



from bidict import bidict

DEFAULT_SINGER = "扇宝"
DEFAULT_SINGER_ID = "513singer"


singer2id = bidict(
    {
        "扇宝": "513singer",
        "SING-林嘉慧": "514singer",
        "Rocky": "881singer",
        "杨超越": "ycysinger",
    },
)

id2singer = singer2id.inverse



# mypy: disable-error-code="misc"
from __future__ import annotations

from enum import Enum
from types import SimpleNamespace
from typing import Annotated, Literal

from pydantic import Field, ValidationInfo, field_validator

from libresvip.model.base import BaseModel
from libresvip.model.option_mixins import (
    EnableBreathImportationMixin,
    EnableGenderImportationMixin,
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _

from .enums import AcepLyricsLanguage
from .singers import DEFAULT_SINGER


class StrengthMappingOption(Enum):
    BOTH: Annotated[
        str,
        Field(
            title=_("Both strength and tension"),
            description=_(
                "Map both strength and tension parameters to strength and tension parameters, each with a weight of 50%."
            ),
        ),
    ] = "both"
    ENERGY: Annotated[
        str,
        Field(
            title=_("Only strength"),
            description=_(
                "Map only strength parameters to strength parameters. Tension parameters will remain unparameterized."
            ),
        ),
    ] = "energy"
    TENSION: Annotated[
        str,
        Field(
            title=_("Only tension"),
            description=_(
                "Map only tension parameters to tension parameters. Strength parameters will remain unparameterized."
            ),
        ),
    ] = "tension"


NormalizationMethod = SimpleNamespace(
    NONE="none",
    ZSCORE="zscore",
    MINMAX="minmax",
)


class NormalizationArgument(BaseModel):
    normalize_method: Literal["none", "zscore", "minmax"] = NormalizationMethod.NONE
    lower_threshold: float = Field(default=0, ge=0.0, le=10.0)
    upper_threshold: float = Field(default=10, ge=0.0, le=10.0)
    scale: float = Field(default=0, ge=-1.0, le=1.0)
    bias: float = Field(default=0, ge=-1.0, le=1.0)

    @classmethod
    def default_repr(cls) -> str:
        fields = cls.model_fields
        default_strs = [str(field.default) for field in fields.values()]
        return ",".join(default_strs)

    @classmethod
    def from_str(cls, v: str) -> NormalizationArgument:
        fields = cls.model_fields
        try:
            obj = cls(**dict(zip(fields.keys(), v.split(","))))
        except Exception:
            obj = cls()
        return obj

    @property
    def enabled(self) -> bool:
        return self.normalize_method != "none"


class InputOptions(
    EnableBreathImportationMixin,
    EnableGenderImportationMixin,
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    BaseModel,
):
    keep_all_pronunciation: bool = Field(
        default=False,
        title=_("Keep all pronunciation information"),
        description=_(
            "ACE Studio will add pronunciation to all notes. This is a redundant data for most Chinese characters, so by default only modified pronunciation will be kept. When this option is turned on, all pronunciation information will be kept unconditionally, but it may produce a larger file size. When the source file contains non-Chinese singing data, please turn on this option."
        ),
    )
    import_tension: bool = Field(
        default=True,
        title=_("Import tension envelope"),
        description=_(
            "When turned on, the tension envelope will be mapped to the tension channel of the intermediate model."
        ),
    )
    import_energy: bool = Field(
        default=True,
        title=_("Import energy envelope"),
        description=_(
            "When turned on, the energy envelope will be mapped to the energy channel of the intermediate model."
        ),
    )
    energy_coefficient: float = Field(
        default=0.5,
        title=_("Strength-volume mapping coefficient"),
        description=_(
            "Since the strength parameter of ACE Studio has a significant impact on the volume, this option is provided to control the mapping coefficient of the strength envelope. The strength envelope will be multiplied by the value of this option as a whole and then mapped to the volume channel of the intermediate model, and the remaining part will be mapped to the strength channel. This option accepts values in the range of 0~1.0."
        ),
    )
    curve_sample_interval: int = Field(
        default=5,
        title=_("Interval of curve sampling"),
        description=_(
            "The unit is tick (the length of a quarter note is 480 ticks). By default, the parameter points in the acep file are stored in a very dense manner. Increasing the sampling interval appropriately will not cause much loss of accuracy, but it can get a smaller file size. Setting to 0 or a negative value means keeping the original sampling interval."
        ),
    )
    breath_normalization: NormalizationArgument = Field(
        default_factory=NormalizationArgument,
        title=_("Breath parameter normalization"),
        description=_("""This option is an advanced option. After enabling this option, the breath parameters will be merged with the breath envelope after being transformed by the normalization. It is recommended to enable this option after fixing all parameters with a fixed brush. This option needs to set 5 values, separated by "," from each other:
(1) Normalization method: none means to turn off this option, zscore means to perform Z-Score normalization on the parameter points, and minmax means to Min-Max normalize the parameter points to the [-1.0, 1.0] interval.
(2) Lower threshold: a real number in the range of 0~10.0, the parameter points lower than this value will not participate in the normalization.
(3) Upper threshold: a real number in the range of 0~10.0, the parameter points higher than this value will not participate in the normalization.
(4) Scaling factor: a real number in the range of -1.0~1.0, the normalized parameter value will be multiplied by this value.
(5) Bias: a real number in the range of -1.0~1.0, the normalized and scaled parameter value will be added to this value."""),
    )
    tension_normalization: NormalizationArgument = Field(
        default_factory=NormalizationArgument,
        title=_("Tension parameter normalization"),
        description=_("""This option is an advanced option. After enabling this option, the tension parameters will be merged with the tension envelope after being transformed by the normalization. It is recommended to enable this option after fixing all parameters with a fixed brush. This option needs to set 5 values, separated by "," from each other:
(1) Normalization method: none means to turn off this option, zscore means to perform Z-Score normalization on the parameter points, and minmax means to Min-Max normalize the parameter points to the [-1.0, 1.0] interval.
(2) Lower threshold: a real number in the range of 0~10.0, the parameter points lower than this value will not participate in the normalization.
(3) Upper threshold: a real number in the range of 0~10.0, the parameter points higher than this value will not participate in the normalization.
(4) Scaling factor: a real number in the range of -1.0~1.0, the normalized parameter value will be multiplied by this value.
(5) Bias: a real number in the range of -1.0~1.0, the normalized and scaled parameter value will be added to this value."""),
    )
    energy_normalization: NormalizationArgument = Field(
        default_factory=NormalizationArgument,
        title=_("Energy parameter normalization"),
        description=_("""This option is an advanced option. After enabling this option, the energy parameters will be merged with the energy envelope after being transformed by the normalization. It is recommended to enable this option after fixing all parameters with a fixed brush. This option needs to set 5 values, separated by "," from each other:
(1) Normalization method: none means to turn off this option, zscore means to perform Z-Score normalization on the parameter points, and minmax means to Min-Max normalize the parameter points to the [-1.0, 1.0] interval.
(2) Lower threshold: a real number in the range of 0~10.0, the parameter points lower than this value will not participate in the normalization.
(3) Upper threshold: a real number in the range of 0~10.0, the parameter points higher than this value will not participate in the normalization.
(4) Scaling factor: a real number in the range of -1.0~1.0, the normalized parameter value will be multiplied by this value.
(5) Bias: a real number in the range of -1.0~1.0, the normalized and scaled parameter value will be added to this value."""),
    )

    @field_validator(
        "breath_normalization",
        "tension_normalization",
        "energy_normalization",
        mode="before",
    )
    @classmethod
    def _validate_normalization_argument(
        cls, v: str | dict[str, str | float], _info: ValidationInfo
    ) -> NormalizationArgument:
        if isinstance(v, str):
            v = NormalizationArgument.from_str(v).model_dump()
        return NormalizationArgument.model_validate(v)


class OutputOptions(BaseModel):
    singer: str = Field(
        default=DEFAULT_SINGER,
        title=_("Default singer"),
        description=_("Please input the complete and correct singer name"),
    )
    breath: int = Field(
        default=600,
        title=_("Default breath length (ms)"),
        description=_(
            "This option is used to set the default breath length when the breath mark is converted to a breath parameter. The actual breath length may be less than the default value due to the small gap between notes; some notes may be shortened due to the insertion of breath marks. Setting to 0 or a negative value means ignoring all breath marks."
        ),
    )
    map_strength_info: StrengthMappingOption = Field(
        default=StrengthMappingOption.BOTH,
        title=_("Map strength and tension parameters to"),
        description=_(
            "ACE Studio has both strength and tension parameters, both of which can affect the strength of the singing. This option is used to set the mapping target of the strength parameter."
        ),
    )
    split_threshold: int = Field(
        default=1,
        title=_("Threshold for splitting"),
        description=_(
            "When the distance between notes exceeds the set value, they will be split into different segments (patterns) for subsequent editing. The threshold unit is the value of a quarter note, and the default is 1, which means that when the distance between notes exceeds 1 quarter notes (480 ticks), they will be split. If you don't want to split at all, please set this option to 0 or a negative value."
        ),
    )
    lyric_language: AcepLyricsLanguage = Field(
        default=AcepLyricsLanguage.CHINESE,
        title=_("Lyrics language"),
        description=_(
            "ACE Studio supports four languages of lyrics. This option is used to set the language of lyrics."
        ),
    )
    default_consonant_length: float = Field(
        default=0.0,
        title=_("Default consonant length (secs)"),
        description=_("Set default consonant length for notes if not specified"),
    )



import dataclasses

from libresvip.core.constants import KEY_IN_OCTAVE
from libresvip.core.lyric_phoneme.japanese import is_kana, is_romaji
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.translation import gettext_lazy as _

from .constants import DEFAULT_PHONEME, OCTAVE_OFFSET, TICK_RATE
from .model import (
    VoiSonaMobileBeatItem,
    VoiSonaMobileNoteItem,
    VoiSonaMobileParameterItem,
    VoiSonaMobileParametersItem,
    VoiSonaMobilePointData,
    VoiSonaMobileProject,
    VoiSonaMobileScoreItem,
    VoiSonaMobileSinger,
    VoiSonaMobileSongItem,
    VoiSonaMobileSoundItem,
    VoiSonaMobileTempoItem,
    VoiSonaMobileTimeItem,
)
from .options import OutputOptions
from .voisona_mobile_pitch import generate_for_voisona


@dataclasses.dataclass
class VoiSonaMobileGenerator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> VoiSonaMobileProject:
        if self.options.track_index < 0:
            first_singing_track = next(
                (track for track in project.track_list if isinstance(track, SingingTrack)),
                None,
            )
        else:
            first_singing_track = project.track_list[self.options.track_index]
            assert isinstance(first_singing_track, SingingTrack)
        voisona_project = VoiSonaMobileProject()
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = int(project.time_signature_list[0].bar_length())
        default_time_signatures = self.generate_time_signatures(project.time_signature_list)
        default_tempos = self.generate_tempos(project.song_tempo_list)
        if first_singing_track is not None:
            singing_track = VoiSonaMobileSinger()
            song_item = VoiSonaMobileSongItem(
                beat=[default_time_signatures],
                tempo=[default_tempos],
                score=[
                    VoiSonaMobileScoreItem(note=self.generate_notes(first_singing_track.note_list))
                ],
            )
            singing_track.mobile_singer_data.state_information.song = [song_item]
            voisona_project.mobile_singer.append(singing_track)
            if log_f0 := self.generate_pitch(
                first_singing_track.edited_params.pitch, project.song_tempo_list
            ):
                singing_track.mobile_singer_data.state_information.parameter = [
                    VoiSonaMobileParametersItem(log_f0=[log_f0])
                ]
        return voisona_project

    def generate_tempos(self, tempos: list[SongTempo]) -> VoiSonaMobileTempoItem:
        return VoiSonaMobileTempoItem(
            sound=[
                VoiSonaMobileSoundItem(
                    clock=round(tempo.position * TICK_RATE) if i else 0,
                    tempo=tempo.bpm,
                )
                for i, tempo in enumerate(tempos)
            ]
        )

    def generate_time_signatures(
        self, time_signatures: list[TimeSignature]
    ) -> VoiSonaMobileBeatItem:
        beat = VoiSonaMobileBeatItem(
            time=[
                VoiSonaMobileTimeItem(
                    clock=0,
                    beats=time_signatures[0].numerator,
                    beat_type=time_signatures[0].denominator,
                )
            ]
        )
        tick = 0.0
        prev_time_signature = time_signatures[0]
        for time_signature in time_signatures[1:]:
            if time_signature.bar_index > prev_time_signature.bar_index:
                tick += (
                    time_signature.bar_index - prev_time_signature.bar_index
                ) * prev_time_signature.bar_length()
            beat.time.append(
                VoiSonaMobileTimeItem(
                    clock=int(tick * TICK_RATE),
                    beats=time_signature.numerator,
                    beat_type=time_signature.denominator,
                )
            )
            prev_time_signature = time_signature
        return beat

    def generate_notes(self, notes: list[Note]) -> list[VoiSonaMobileNoteItem]:
        voisona_notes = []
        for note in notes:
            lyric = note.lyric
            phoneme = ""
            if note.pronunciation:
                phoneme = note.pronunciation
            elif not is_kana(lyric) and not is_romaji(lyric):
                phoneme = DEFAULT_PHONEME
                msg_prefix = _("Unsupported lyric: ")
                show_warning(f"{msg_prefix} {lyric}")
            voisona_notes.append(
                VoiSonaMobileNoteItem(
                    clock=int(note.start_pos * TICK_RATE),
                    duration=int(note.length * TICK_RATE),
                    lyric=note.lyric,
                    pitch_octave=note.key_number // KEY_IN_OCTAVE + OCTAVE_OFFSET,
                    pitch_step=note.key_number % KEY_IN_OCTAVE,
                    syllabic=0,
                    phoneme=phoneme,
                )
            )
        return voisona_notes

    def generate_pitch(
        self, pitch: ParamCurve, tempo_list: list[SongTempo]
    ) -> VoiSonaMobileParameterItem | None:
        if (data := generate_for_voisona(pitch, tempo_list, self.first_bar_length)) is not None:
            return VoiSonaMobileParameterItem(
                length=data.length,
                data=[
                    VoiSonaMobilePointData(
                        index=each.idx,
                        repeat=each.repeat,
                        value=each.value,
                    )
                    for each in data.events
                ],
            )



from pydantic import BaseModel

from libresvip.model.option_mixins import (
    EnablePitchImportationMixin,
    SelectSingleTrackMixin,
)


class InputOptions(
    EnablePitchImportationMixin,
    BaseModel,
):
    pass


class OutputOptions(SelectSingleTrackMixin, BaseModel):
    pass



import dataclasses

import pypinyin

from libresvip.model.base import (
    Note,
    ParamCurve,
    Project,
    SingingTrack,
)

from .model import Y77Note, Y77Project
from .options import OutputOptions


@dataclasses.dataclass
class Y77Generator:
    options: OutputOptions
    first_bar_length: int = 0

    def generate_project(self, project: Project) -> Y77Project:
        if self.options.track_index < 0:
            first_singing_track = next(
                (track for track in project.track_list if isinstance(track, SingingTrack)),
                None,
            )
        else:
            first_singing_track = project.track_list[self.options.track_index]
            assert isinstance(first_singing_track, SingingTrack)
        self.first_bar_length = int(project.time_signature_list[0].bar_length())
        y77_project = Y77Project(
            bars=100,
            bpm=project.song_tempo_list[0].bpm,
            bbar=project.time_signature_list[0].numerator,
            bbeat=project.time_signature_list[0].denominator,
        )
        if first_singing_track is not None:
            y77_project.notes = self.generate_notes(first_singing_track)
            y77_project.nnote = len(y77_project.notes)
        return y77_project

    def generate_notes(self, singing_track: SingingTrack) -> list[Y77Note]:
        return [
            Y77Note(
                lyric=note.lyric,
                start=round(note.start_pos / 30),
                length=round(note.length / 30),
                pitch=88 - note.key_number,
                py=note.pronunciation or " ".join(pypinyin.lazy_pinyin(note.lyric)),
                pit=self.generate_pitch(singing_track.edited_params.pitch, note),
            )
            for note in singing_track.note_list
        ]

    def generate_pitch(self, pitch_param_curve: ParamCurve, note: Note) -> list[float]:
        sample_time_list = [
            note.start_pos + self.first_bar_length + int((note.length / 500.0) * i)
            for i in range(500)
        ]
        pitch_param_in_note = [
            p
            for p in pitch_param_curve.points.root
            if p.x >= note.start_pos + self.first_bar_length
            and p.x <= note.end_pos + self.first_bar_length
        ]

        pitch_param_time_in_note = dict(pitch_param_in_note)

        y77_pitch_param: list[int | float] = []
        for sample_time in sample_time_list:
            if (pitch := pitch_param_time_in_note.get(sample_time)) is None:
                distance = -1
                value = 50.0

                for point in pitch_param_in_note:
                    if distance > abs(point.x - sample_time) or distance == -1:
                        distance = abs(point.x - sample_time)
                        value = 50 + (
                            0 if point.y == -100 else (point.y - note.key_number * 100) / 2
                        )

                y77_pitch_param.append(value)

            elif pitch == -100:
                y77_pitch_param.append(50)
            else:
                y77_pitch_param.append(50 + (pitch - note.key_number * 100) / 2)
        buffer = []
        previous_node = y77_pitch_param[0]
        previous_node_index = 0
        for i in range(len(y77_pitch_param)):
            if y77_pitch_param[i] == previous_node:
                buffer.append(y77_pitch_param[i])
            else:
                for j in range(len(buffer)):
                    y77_pitch_param[previous_node_index + j] = previous_node + j * (
                        y77_pitch_param[i] - buffer[j]
                    ) / len(buffer)
                buffer.clear()

            if y77_pitch_param[i] != previous_node:
                previous_node_index = i
                previous_node = y77_pitch_param[i]

        return y77_pitch_param



from __future__ import annotations

import copy
import dataclasses
import functools
import gettext
import inspect
import sys
from importlib.machinery import (
    PathFinder,
    SourceFileLoader,
    all_suffixes,
)
from importlib.resources import files
from typing import TYPE_CHECKING, Generic, TypeGuard, cast

from loguru import logger
from typing_extensions import TypeVar

from libresvip.core.config import get_ui_settings, settings
from libresvip.core.constants import app_dir, pkg_dir, res_dir
from libresvip.utils.module_loading import import_module

from .base import BasePlugin_co, MiddlewareBase, SVSConverterBase
from .meta_info import (
    BasePluginInfo,
    FormatProviderPluginInfo,
    MiddlewarePluginInfo,
)

if TYPE_CHECKING:
    import pathlib
    from importlib.abc import MetaPathFinder
    from importlib.machinery import ModuleSpec
    from types import ModuleType

    from libresvip.core.compat import Traversable


PluginInfo_co = TypeVar("PluginInfo_co", bound="BasePluginInfo", covariant=True)


@dataclasses.dataclass
class BasePluginManager(Generic[BasePlugin_co, PluginInfo_co]):
    info_extension: str
    plugin_base: type[BasePlugin_co]
    plugin_info_class: type[PluginInfo_co]
    plugin_namespace: str
    install_path: pathlib.Path
    plugin_places: list[Traversable]
    plugin_registry: dict[str, PluginInfo_co] = dataclasses.field(default_factory=dict)
    _candidates: list[tuple[Traversable, PluginInfo_co]] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        sys.meta_path.append(cast("MetaPathFinder", self))

    @functools.cached_property
    def lib_suffixes(self) -> list[str]:
        return all_suffixes()

    def find_spec(
        self,
        fullname: str,
        path: list[str] | None,
        target: ModuleType | None = None,
    ) -> ModuleSpec | None:
        if not fullname.startswith(self.plugin_namespace) or not path:
            return None
        path = [str(path) for path in self.plugin_places]

        if (
            spec := PathFinder.find_spec(fullname, path, target)
        ) is not None and spec.loader is not None:
            spec.loader = SourceFileLoader(spec.loader.name, spec.loader.path)  # type: ignore[attr-defined]
            return spec

    def is_plugin(self, member: object) -> TypeGuard[type[BasePlugin_co]]:
        return (
            inspect.isclass(member)
            and not inspect.isabstract(member)
            and issubclass(member, self.plugin_base)
        )

    def scan_candidates(self) -> None:
        self._candidates.clear()
        _discovered: set[str] = set()
        for dir_path in self.plugin_places:
            # first of all, is it a directory :)
            if not dir_path.is_dir():
                logger.debug(f"{self.__class__.__name__} skips {dir_path} (not a directory)")
                continue
            # iteratively walks through the directory
            for child_path in dir_path.iterdir():
                if not child_path.is_dir():
                    continue
                for file_path in child_path.iterdir():
                    if not file_path.is_file() or not file_path.name.endswith(
                        f".{self.info_extension}"
                    ):
                        continue
                    if (candidate_infofile := str(file_path)) in _discovered:
                        logger.debug(f"{candidate_infofile} rejected because already discovered")
                        continue
                    logger.debug(
                        f"{self.__class__.__name__} found a candidate:\n    {candidate_infofile}"
                    )
                    if (plugin_info := self.plugin_info_class.load(file_path)) is None:
                        logger.debug(f"Plugin candidate '{candidate_infofile}'  rejected")
                        continue  # we consider this was the good strategy to use for: it failed -> not a plugin -> don't try another strategy
                    if entry_suffix := next(
                        (
                            suffix
                            for suffix in self.lib_suffixes
                            if (
                                (
                                    candidate_filepath := (
                                        child_path / f"{plugin_info.module}{suffix}"
                                    )
                                ).is_file()
                            )
                        ),
                        None,
                    ):
                        _discovered.add(
                            str(
                                child_path / f"{plugin_info.module}{entry_suffix}",
                            )
                        )
                    else:
                        logger.error(
                            f"Plugin candidate rejected: cannot find the file or directory module for '{candidate_infofile}'",
                        )
                        break
                    self._candidates.append((candidate_filepath, plugin_info))
                    # finally the candidate_infofile must not be discovered again
                    _discovered.add(candidate_infofile)

    def import_plugins(self, reload: bool = False) -> None:
        if reload:
            self.plugin_registry.clear()
        self.scan_candidates()
        for candidate_filepath, plugin_info in self._candidates:
            # make sure to attribute a unique module name to the one
            # that is about to be loaded
            plugin_module_name = f"{self.plugin_namespace}.{plugin_info.identifier}"
            if (
                plugin_info.identifier in self.plugin_registry and not reload
            ) or plugin_info.identifier in settings.disabled_plugins:
                logger.debug(f"Skipped plugin: {plugin_info.identifier}")
                continue
            try:
                candidate_module = import_module(plugin_module_name, candidate_filepath, reload)
            except Exception as e:
                logger.exception(e)
                logger.error(
                    f"Unable to import plugin: {candidate_filepath}",
                )
                continue

            try:
                plugin_cls_name, plugin_cls = inspect.getmembers(candidate_module, self.is_plugin)[
                    0
                ]
                plugin_info.plugin_object = plugin_cls()
                self.plugin_registry[plugin_info.identifier] = plugin_info
            except Exception:
                logger.error(f"Unable to create plugin object: {candidate_filepath}")
                continue  # If it didn't work once it wont again


class ConverterPluginManager(BasePluginManager[SVSConverterBase, FormatProviderPluginInfo]):
    pass


class MiddlewareManager(BasePluginManager[MiddlewareBase, MiddlewarePluginInfo]):
    pass


plugin_manager = ConverterPluginManager(
    info_extension="yapsy-plugin",
    plugin_base=SVSConverterBase,  # type: ignore [type-abstract]
    plugin_info_class=FormatProviderPluginInfo,
    plugin_places=[pkg_dir / "plugins", app_dir.user_config_path / "plugins"],
    plugin_namespace="libresvip.plugins",
    install_path=app_dir.user_config_path / "plugins",
)
plugin_manager.import_plugins()
middleware_manager = MiddlewareManager(
    info_extension="yapsy-plugin",
    plugin_base=MiddlewareBase,  # type: ignore [type-abstract]
    plugin_info_class=MiddlewarePluginInfo,
    plugin_places=[
        pkg_dir / "middlewares",
        app_dir.user_config_path / "middlewares",
    ],
    plugin_namespace="libresvip.middlewares",
    install_path=app_dir.user_config_path / "middlewares",
)
middleware_manager.import_plugins()


def merge_translation(
    ori_translation: gettext.NullTranslations,
    resource_dir: Traversable,
    lang: str,
) -> gettext.NullTranslations:
    msg_dir = resource_dir / "locales" / lang / "LC_MESSAGES"
    if msg_dir.is_dir():
        for child_file in msg_dir.iterdir():
            if child_file.name.endswith(".mo"):
                with child_file.open("rb") as fp:
                    child_translation = gettext.GNUTranslations(fp)
                    new_translation = copy.copy(ori_translation)
                    new_translation.add_fallback(child_translation)
                    return new_translation
    else:
        logger.debug(f"No translation file found in {msg_dir}")
    return ori_translation


def get_translation(lang: str | None = None) -> gettext.NullTranslations:
    if lang is None:
        lang = get_ui_settings().language.value
    translation = gettext.NullTranslations()
    translation = merge_translation(translation, res_dir, lang)
    for manager in [plugin_manager, middleware_manager]:
        for plugin_id in manager.plugin_registry:
            plugin_base_dir = files(f"{manager.plugin_namespace}.{plugin_id}")
            translation = merge_translation(translation, plugin_base_dir, lang)
    return translation



from __future__ import annotations

from typing import TYPE_CHECKING

import proto
from google.protobuf import any_pb2

if TYPE_CHECKING:
    from collections.abc import MutableSequence

__protobuf__ = proto.module(
    package="xstudio.proto",
    manifest={
        "Svip3TrackType",
        "Svip3PatternType",
        "Svip3NoteLengthValidateTag",
        "Svip3Project",
        "Svip3SongTempo",
        "Svip3SongBeat",
        "Svip3SongTone",
        "Svip3BeatSize",
        "Svip3Master",
        "Svip3SingingTrack",
        "Svip3AudioTrack",
        "Svip3AudioPattern",
        "Svip3SingingPattern",
        "Svip3Vibrato",
        "Svip3Note",
        "Svip3VibratoStyle",
        "Svip3LineParamNode",
    },
)


class Svip3TrackType(proto.Enum):
    Track_None = 0
    Audio_Track = 1
    Singing_Track = 2


class Svip3PatternType(proto.Enum):
    Pattern_None = 0
    Audio_Pattern = 1
    Singing_Pattern = 2


class Svip3NoteLengthValidateTag(proto.Enum):
    NONE = 0
    TOO_LONG = 1
    TOO_SHORT = 2


class Svip3SongMark(proto.Message):
    pos: int = proto.Field(
        proto.INT32,
        number=1,
    )
    mark: str = proto.Field(
        proto.STRING,
        number=2,
    )


class Svip3Project(proto.Message):
    project_file_path: str = proto.Field(
        proto.STRING,
        number=1,
    )
    version: str = proto.Field(
        proto.STRING,
        number=2,
    )
    duration: int = proto.Field(
        proto.INT32,
        number=3,
    )
    tempo_list: MutableSequence[Svip3SongTempo] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message="Svip3SongTempo",
    )
    beat_list: MutableSequence[Svip3SongBeat] = proto.RepeatedField(
        proto.MESSAGE,
        number=5,
        message="Svip3SongBeat",
    )
    track_list: MutableSequence[any_pb2.Any] = proto.RepeatedField(
        proto.MESSAGE,
        number=6,
        message=any_pb2.Any,
    )
    master: Svip3Master = proto.Field(
        proto.MESSAGE,
        number=7,
        message="Svip3Master",
    )
    current_tone: str = proto.Field(
        proto.STRING,
        number=8,
    )
    piano_cells: int = proto.Field(
        proto.INT32,
        number=9,
    )
    loop_start: int = proto.Field(
        proto.INT32,
        number=10,
    )
    loop_end: int = proto.Field(
        proto.INT32,
        number=11,
    )
    is_open_adsorb: bool = proto.Field(
        proto.BOOL,
        number=12,
    )
    params_version: int = proto.Field(
        proto.INT32,
        number=13,
    )
    tone_list: MutableSequence[Svip3SongTone] = proto.RepeatedField(
        proto.MESSAGE,
        number=14,
        message="Svip3SongTone",
    )
    is_triplets: bool = proto.Field(
        proto.BOOL,
        number=15,
    )
    is_loop: bool = proto.Field(
        proto.BOOL,
        number=16,
    )
    is_last_play: bool = proto.Field(
        proto.BOOL,
        number=17,
    )
    mark_list: MutableSequence[Svip3SongMark] = proto.RepeatedField(
        proto.MESSAGE,
        number=18,
        message="Svip3SongMark",
    )


class Svip3SongTempo(proto.Message):
    pos: int = proto.Field(
        proto.INT32,
        number=1,
    )
    tempo: int = proto.Field(
        proto.INT32,
        number=2,
    )


class Svip3SongBeat(proto.Message):
    pos: int = proto.Field(
        proto.INT32,
        number=1,
    )
    beat_size: Svip3BeatSize = proto.Field(
        proto.MESSAGE,
        number=2,
        message="Svip3BeatSize",
    )


class Svip3SongTone(proto.Message):
    pos: int = proto.Field(
        proto.INT32,
        number=1,
    )
    tone: str = proto.Field(
        proto.STRING,
        number=2,
    )


class Svip3BeatSize(proto.Message):
    numerator: int = proto.Field(
        proto.INT32,
        number=1,
    )
    denominator: int = proto.Field(
        proto.INT32,
        number=2,
    )


class Svip3Master(proto.Message):
    volume: float = proto.Field(
        proto.FLOAT,
        number=1,
    )


class Svip3SingingTrack(proto.Message):
    volume: float = proto.Field(
        proto.FLOAT,
        number=1,
    )
    pan: float = proto.Field(
        proto.FLOAT,
        number=2,
    )
    mute: bool = proto.Field(
        proto.BOOL,
        number=3,
    )
    name: str = proto.Field(
        proto.STRING,
        number=4,
    )
    solo: bool = proto.Field(
        proto.BOOL,
        number=5,
    )
    color: str = proto.Field(
        proto.STRING,
        number=6,
    )
    type_: Svip3TrackType = proto.Field(
        proto.ENUM,
        number=7,
        enum="Svip3TrackType",
    )
    pattern_list: MutableSequence[Svip3SingingPattern] = proto.RepeatedField(
        proto.MESSAGE,
        number=8,
        message="Svip3SingingPattern",
    )
    ai_singer_id: str = proto.Field(
        proto.STRING,
        number=9,
    )
    is_reverb_open: bool = proto.Field(
        proto.BOOL,
        number=10,
    )
    reverb_type: int = proto.Field(
        proto.INT32,
        number=11,
    )
    reverb_db: float = proto.Field(
        proto.FLOAT,
        number=12,
    )
    is_rap_track: bool = proto.Field(
        proto.BOOL,
        number=13,
    )


class Svip3AudioTrack(proto.Message):
    volume: float = proto.Field(
        proto.FLOAT,
        number=1,
    )
    pan: float = proto.Field(
        proto.FLOAT,
        number=2,
    )
    mute: bool = proto.Field(
        proto.BOOL,
        number=3,
    )
    name: str = proto.Field(
        proto.STRING,
        number=4,
    )
    solo: bool = proto.Field(
        proto.BOOL,
        number=5,
    )
    color: str = proto.Field(
        proto.STRING,
        number=6,
    )
    type_: Svip3TrackType = proto.Field(
        proto.ENUM,
        number=7,
        enum="Svip3TrackType",
    )
    pattern_list: MutableSequence[Svip3AudioPattern] = proto.RepeatedField(
        proto.MESSAGE,
        number=8,
        message="Svip3AudioPattern",
    )


class Svip3AudioPattern(proto.Message):
    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    type_: Svip3PatternType = proto.Field(
        proto.ENUM,
        number=2,
        enum="Svip3PatternType",
    )
    real_pos: int = proto.Field(
        proto.INT32,
        number=3,
    )
    real_dur: int = proto.Field(
        proto.INT32,
        number=4,
    )
    play_pos: int = proto.Field(
        proto.INT32,
        number=5,
    )
    play_dur: int = proto.Field(
        proto.INT32,
        number=6,
    )
    is_mute: bool = proto.Field(
        proto.BOOL,
        number=7,
    )
    audio_file_path: str = proto.Field(
        proto.STRING,
        number=8,
    )
    rising_falling_tone: float = proto.Field(
        proto.FLOAT,
        number=9,
    )


class Svip3SingingPattern(proto.Message):
    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    type_: Svip3PatternType = proto.Field(
        proto.ENUM,
        number=2,
        enum="Svip3PatternType",
    )
    real_pos: int = proto.Field(
        proto.INT32,
        number=3,
    )
    real_dur: int = proto.Field(
        proto.INT32,
        number=4,
    )
    play_pos: int = proto.Field(
        proto.INT32,
        number=5,
    )
    play_dur: int = proto.Field(
        proto.INT32,
        number=6,
    )
    is_mute: bool = proto.Field(
        proto.BOOL,
        number=7,
    )
    note_list: MutableSequence[Svip3Note] = proto.RepeatedField(
        proto.MESSAGE,
        number=8,
        message="Svip3Note",
    )
    edited_pitch_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=9,
        message="Svip3LineParamNode",
    )
    edited_volume_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=10,
        message="Svip3LineParamNode",
    )
    edited_power_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=11,
        message="Svip3LineParamNode",
    )
    merge_pitch_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=12,
        message="Svip3LineParamNode",
    )
    merge_power_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=13,
        message="Svip3LineParamNode",
    )
    edited_spec_trans_coef_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=14,
        message="Svip3LineParamNode",
    )
    edited_ap_coef_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=15,
        message="Svip3LineParamNode",
    )
    edited_energy_value_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=16,
        message="Svip3LineParamNode",
    )
    merge_energy_value_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=17,
        message="Svip3LineParamNode",
    )
    rap_type: int = proto.Field(
        proto.INT32,
        number=18,
    )
    singing_method_id: str = proto.Field(
        proto.STRING,
        number=19,
    )

    @property
    def pos(self) -> int:
        return self.real_pos + self.play_pos


class Svip3Vibrato(proto.Message):
    frequency: float = proto.Field(
        proto.FLOAT,
        number=1,
    )
    amplitude: float = proto.Field(
        proto.FLOAT,
        number=2,
    )
    phase: float = proto.Field(
        proto.FLOAT,
        number=3,
    )
    start: float = proto.Field(
        proto.FLOAT,
        number=4,
    )
    end: float = proto.Field(
        proto.FLOAT,
        number=5,
    )
    attack_x: float = proto.Field(
        proto.FLOAT,
        number=6,
    )
    attack_y: float = proto.Field(
        proto.FLOAT,
        number=7,
    )
    release_x: float = proto.Field(
        proto.FLOAT,
        number=8,
    )
    release_y: float = proto.Field(
        proto.FLOAT,
        number=9,
    )


class Svip3Note(proto.Message):
    start_pos: int = proto.Field(
        proto.INT32,
        number=1,
    )
    width_pos: int = proto.Field(
        proto.INT32,
        number=2,
    )
    key_index: int = proto.Field(
        proto.INT32,
        number=3,
    )
    lyric: str = proto.Field(
        proto.STRING,
        number=4,
    )
    pronouncing: str = proto.Field(
        proto.STRING,
        number=5,
    )
    consonant_len: int = proto.Field(
        proto.INT32,
        number=6,
    )
    has_consonant: bool = proto.Field(
        proto.BOOL,
        number=7,
    )
    user_consonant_len: int = proto.Field(
        proto.INT32,
        number=8,
    )
    sp_len: int = proto.Field(
        proto.INT32,
        number=9,
    )
    sil_len: int = proto.Field(
        proto.INT32,
        number=10,
    )
    length_validate_tag: Svip3NoteLengthValidateTag = proto.Field(
        proto.ENUM,
        number=11,
        enum="Svip3NoteLengthValidateTag",
    )
    vibrato: Svip3Vibrato = proto.Field(
        proto.MESSAGE,
        number=12,
        message="Svip3Vibrato",
    )
    user_sp_len: int = proto.Field(
        proto.INT32,
        number=13,
    )
    tone: int = proto.Field(
        proto.INT32,
        number=14,
    )
    left_pitch: float = proto.Field(
        proto.FLOAT,
        number=15,
    )
    right_pitch: float = proto.Field(
        proto.FLOAT,
        number=16,
    )


class Svip3VibratoStyle(proto.Message):
    is_anti_phase: bool = proto.Field(
        proto.BOOL,
        number=1,
    )
    amp_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=2,
        message="Svip3LineParamNode",
    )
    freq_line: MutableSequence[Svip3LineParamNode] = proto.RepeatedField(
        proto.MESSAGE,
        number=3,
        message="Svip3LineParamNode",
    )


class Svip3LineParamNode(proto.Message):
    pos: int = proto.Field(
        proto.SINT32,
        number=1,
    )
    value: float = proto.Field(
        proto.FLOAT,
        number=2,
    )



from decimal import Decimal

from xsdata.models.datatype import XmlTime
from xsdata_pydantic.fields import field

from libresvip.model.base import BaseModel


class CeVIOAuthor(BaseModel):
    class Meta:
        name = "Author"

    version: str | None = field(
        default="3.2.21.2",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )


class CeVIONoData(BaseModel):
    class Meta:
        name = "NoData"

    index: int | None = field(
        default=None,
        metadata={
            "name": "Index",
            "type": "Attribute",
        },
    )
    repeat: int | None = field(
        default=None,
        metadata={
            "name": "Repeat",
            "type": "Attribute",
        },
    )


class CeVIOData(CeVIONoData):
    class Meta:
        name = "Data"

    value: float | int | Decimal | None = field(default=None)


class CeVIODictExtension(BaseModel):
    class Meta:
        name = "Extension"

    version: str | None = field(
        default="1.0.0",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )

    language: str | None = field(
        default="English",
        metadata={
            "name": "Language",
            "type": "Attribute",
        },
    )


class CeVIODictionary(BaseModel):
    class Meta:
        name = "Dictionary"

    version: str | None = field(
        default="1.0.0",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )
    extension: CeVIODictExtension | None = field(
        default=None,
        metadata={
            "name": "Extension",
            "type": "Element",
        },
    )


class CeVIODynamics(BaseModel):
    class Meta:
        name = "Dynamics"

    clock: int | None = field(
        default=None,
        metadata={
            "name": "Clock",
            "type": "Attribute",
        },
    )
    value: int | None = field(
        default=None,
        metadata={
            "name": "Value",
            "type": "Attribute",
        },
    )


class CeVIOTalkExtension(BaseModel):
    class Meta:
        name = "Extension"

    vertical_ratio: str | None = field(
        default=None,
        metadata={
            "name": "VerticalRatio",
            "type": "Attribute",
        },
    )


class CeVIOGroup(BaseModel):
    class Meta:
        name = "Group"

    version: str | None = field(
        default="1.0",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )
    group_id: str | None = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )
    category: str = field(
        default="SingerSong",
        metadata={
            "name": "Category",
            "type": "Attribute",
        },
    )
    name: str | None = field(
        default=None,
        metadata={
            "name": "Name",
            "type": "Attribute",
        },
    )
    color: str | None = field(
        default=None,
        metadata={
            "name": "Color",
            "type": "Attribute",
        },
    )
    volume: float | None = field(
        default=None,
        metadata={
            "name": "Volume",
            "type": "Attribute",
        },
    )
    pan: float | None = field(
        default=None,
        metadata={
            "name": "Pan",
            "type": "Attribute",
        },
    )
    is_solo: bool | None = field(
        default=False,
        metadata={
            "name": "IsSolo",
            "type": "Attribute",
        },
    )
    is_muted: bool | None = field(
        default=False,
        metadata={
            "name": "IsMuted",
            "type": "Attribute",
        },
    )
    cast_id: str | None = field(
        default=None,
        metadata={
            "name": "CastId",
            "type": "Attribute",
        },
    )
    language: str | None = field(
        default="Japanese",
        metadata={
            "name": "Language",
            "type": "Attribute",
        },
    )
    snapshot: str | None = field(
        default=None,
        metadata={
            "name": "SnapShot",
            "type": "Attribute",
        },
    )


class CeVIOKey(BaseModel):
    class Meta:
        name = "Key"

    clock: int | None = field(
        default=None,
        metadata={
            "name": "Clock",
            "type": "Attribute",
        },
    )
    fifths: int | None = field(
        default=None,
        metadata={
            "name": "Fifths",
            "type": "Attribute",
        },
    )
    mode: int | None = field(
        default=None,
        metadata={
            "name": "Mode",
            "type": "Attribute",
        },
    )


class CeVIONote(BaseModel):
    class Meta:
        name = "Note"

    clock: int | None = field(
        default=None,
        metadata={
            "name": "Clock",
            "type": "Attribute",
        },
    )
    pitch_step: int | None = field(
        default=None,
        metadata={
            "name": "PitchStep",
            "type": "Attribute",
        },
    )
    pitch_octave: int | None = field(
        default=None,
        metadata={
            "name": "PitchOctave",
            "type": "Attribute",
        },
    )
    duration: int | None = field(
        default=None,
        metadata={
            "name": "Duration",
            "type": "Attribute",
        },
    )
    lyric: str | None = field(
        default=None,
        metadata={
            "name": "Lyric",
            "type": "Attribute",
        },
    )
    phonetic: str | None = field(
        default=None,
        metadata={
            "name": "Phonetic",
            "type": "Attribute",
        },
    )
    do_re_mi: bool | None = field(
        default=None,
        metadata={
            "name": "DoReMi",
            "type": "Attribute",
        },
    )
    staccato: bool | None = field(
        default=None,
        metadata={
            "name": "Staccato",
            "type": "Attribute",
        },
    )
    slur_start: bool | None = field(
        default=None,
        metadata={
            "name": "SlurStart",
            "type": "Attribute",
        },
    )
    slur_stop: bool | None = field(
        default=None,
        metadata={
            "name": "SlurStop",
            "type": "Attribute",
        },
    )
    syllabic: int | None = field(
        default=None,
        metadata={
            "name": "Syllabic",
            "type": "Attribute",
        },
    )
    accent: bool | None = field(
        default=None,
        metadata={
            "name": "Accent",
            "type": "Attribute",
        },
    )
    breath: bool | None = field(
        default=None,
        metadata={
            "name": "Breath",
            "type": "Attribute",
        },
    )


class CeVIOReferenceState(BaseModel):
    class Meta:
        name = "ReferenceState"

    current: str | None = field(
        default=None,
        metadata={
            "name": "Current",
            "type": "Attribute",
        },
    )
    previous: str | None = field(
        default=None,
        metadata={
            "name": "Previous",
            "type": "Attribute",
        },
    )


class CeVIOSound(BaseModel):
    class Meta:
        name = "Sound"

    clock: int | None = field(
        default=None,
        metadata={
            "name": "Clock",
            "type": "Attribute",
        },
    )
    tempo: float | None = field(
        default=None,
        metadata={
            "name": "Tempo",
            "type": "Attribute",
        },
    )


class CeVIOSoundSetting(BaseModel):
    class Meta:
        name = "SoundSetting"

    rhythm: str | None = field(
        default="4/4",
        metadata={
            "name": "Rhythm",
            "type": "Attribute",
        },
    )
    tempo: float | None = field(
        default=120.0,
        metadata={
            "name": "Tempo",
            "type": "Attribute",
        },
    )
    master_volume: float | None = field(
        default=None,
        metadata={
            "name": "MasterVolume",
            "type": "Attribute",
        },
    )


class CeVIOSoundSource(BaseModel):
    class Meta:
        name = "SoundSource"

    version: str | None = field(
        default="1.0.0",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )
    sound_source_id: str | None = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )
    name: str | None = field(
        default=None,
        metadata={
            "name": "Name",
            "type": "Attribute",
        },
    )


class CeVIOTime(BaseModel):
    class Meta:
        name = "Time"

    clock: int = field(
        metadata={
            "name": "Clock",
            "type": "Attribute",
        },
    )
    beats: int | None = field(
        default=None,
        metadata={
            "name": "Beats",
            "type": "Attribute",
        },
    )
    beat_type: int | None = field(
        default=None,
        metadata={
            "name": "BeatType",
            "type": "Attribute",
        },
    )


class CeVIOViewScale(BaseModel):
    class Meta:
        name = "ViewScale"

    horizontal: int | float | None = field(
        default=None,
        metadata={
            "name": "Horizontal",
            "type": "Attribute",
        },
    )
    vertical: int | Decimal | None = field(
        default=None,
        metadata={
            "name": "Vertical",
            "type": "Attribute",
        },
    )


class CeVIOBeat(BaseModel):
    class Meta:
        name = "Beat"

    time: list[CeVIOTime] = field(
        default_factory=list,
        metadata={
            "name": "Time",
            "type": "Element",
        },
    )


class CeVIOGroups(BaseModel):
    class Meta:
        name = "Groups"

    active_group: str | None = field(
        default=None,
        metadata={
            "name": "ActiveGroup",
            "type": "Attribute",
        },
    )
    group: list[CeVIOGroup] = field(
        default_factory=list,
        metadata={
            "name": "Group",
            "type": "Element",
        },
    )


class CeVIOScore(BaseModel):
    class Meta:
        name = "Score"

    alpha: float | None = field(
        default=None,
        metadata={
            "name": "Alpha",
            "type": "Attribute",
        },
    )
    tune: float | None = field(
        default=None,
        metadata={
            "name": "Tune",
            "type": "Attribute",
        },
    )
    pitch_shift: float | None = field(
        default=None,
        metadata={
            "name": "PitchShift",
            "type": "Attribute",
        },
    )
    pitch_tune: float | None = field(
        default=None,
        metadata={
            "name": "PitchTune",
            "type": "Attribute",
        },
    )
    husky: float | None = field(
        default=None,
        metadata={
            "name": "Husky",
            "type": "Attribute",
        },
    )
    vib_amp: float | None = field(
        default=None,
        metadata={
            "name": "VibAmp",
            "type": "Attribute",
        },
    )
    vib_frq: float | None = field(
        default=None,
        metadata={
            "name": "VibFrq",
            "type": "Attribute",
        },
    )
    emotion0: float | int | None = field(
        default=None,
        metadata={
            "name": "Emotion0",
            "type": "Attribute",
        },
    )
    emotion1: float | int | None = field(
        default=None,
        metadata={
            "name": "Emotion1",
            "type": "Attribute",
        },
    )
    key: list[CeVIOKey] = field(
        default_factory=list,
        metadata={
            "name": "Key",
            "type": "Element",
        },
    )
    dynamics: list[CeVIODynamics] = field(
        default_factory=list,
        metadata={
            "name": "Dynamics",
            "type": "Element",
            "sequential": True,
        },
    )
    note: list[CeVIONote] = field(
        default_factory=list,
        metadata={
            "name": "Note",
            "type": "Element",
            "sequential": True,
        },
    )


class CeVIOSongEditor(BaseModel):
    class Meta:
        name = "SongEditor"

    partition: int | None = field(
        default=None,
        metadata={
            "name": "Partition",
            "type": "Attribute",
        },
    )
    quantize: int | None = field(
        default=None,
        metadata={
            "name": "Quantize",
            "type": "Attribute",
        },
    )
    mode: int | None = field(
        default=None,
        metadata={
            "name": "Mode",
            "type": "Attribute",
        },
    )
    editing_tool: int | None = field(
        default=None,
        metadata={
            "name": "EditingTool",
            "type": "Attribute",
        },
    )
    view_scale: CeVIOViewScale | None = field(
        default=None,
        metadata={
            "name": "ViewScale",
            "type": "Element",
        },
    )
    reference_state: CeVIOReferenceState | None = field(
        default=None,
        metadata={
            "name": "ReferenceState",
            "type": "Element",
        },
    )


class CeVIOSoundSources(BaseModel):
    class Meta:
        name = "SoundSources"

    sound_source: list[CeVIOSoundSource] = field(
        default_factory=list,
        metadata={
            "name": "SoundSource",
            "type": "Element",
        },
    )


class CeVIOTts(BaseModel):
    class Meta:
        name = "TTS"

    version: str | None = field(
        default="3.1.0",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )
    dictionary: CeVIODictionary | None = field(
        default_factory=CeVIODictionary,
        metadata={
            "name": "Dictionary",
            "type": "Element",
        },
    )
    sound_sources: CeVIOSoundSources = field(
        default_factory=CeVIOSoundSources,
        metadata={
            "name": "SoundSources",
            "type": "Element",
        },
    )


class CeVIOTalkEditor(BaseModel):
    class Meta:
        name = "TalkEditor"

    partition: int | None = field(
        default=None,
        metadata={
            "name": "Partition",
            "type": "Attribute",
        },
    )
    extension: CeVIOTalkExtension | None = field(
        default=None,
        metadata={
            "name": "Extension",
            "type": "Element",
        },
    )


class CeVIOTempo(BaseModel):
    class Meta:
        name = "Tempo"

    sound: list[CeVIOSound] = field(
        default_factory=list,
        metadata={
            "name": "Sound",
            "type": "Element",
        },
    )


class CeVIOTimeline(BaseModel):
    class Meta:
        name = "Timeline"

    partition: str | None = field(
        default=None,
        metadata={
            "name": "Partition",
            "type": "Attribute",
        },
    )
    current_position: XmlTime | None = field(
        default=None,
        metadata={
            "name": "CurrentPosition",
            "type": "Attribute",
        },
    )
    start_position: XmlTime | None = field(
        default=None,
        metadata={
            "name": "StartPosition",
            "type": "Attribute",
        },
    )
    end_position: XmlTime | None = field(
        default=None,
        metadata={
            "name": "EndPosition",
            "type": "Attribute",
        },
    )
    view_scale: CeVIOViewScale | None = field(
        default=None,
        metadata={
            "name": "ViewScale",
            "type": "Element",
        },
    )


class CeVIOParameter(BaseModel):
    class Meta:
        name = "Parameter"

    length: int | None = field(
        default=None,
        metadata={
            "name": "Length",
            "type": "Attribute",
        },
    )
    data: list[CeVIOData | float | Decimal | int] = field(
        default_factory=list,
        metadata={
            "name": "Data",
            "type": "Element",
        },
    )
    no_data: list[CeVIONoData] = field(
        default_factory=list,
        metadata={
            "name": "NoData",
            "type": "Element",
        },
    )


class CeVIOParameters(BaseModel):
    class Meta:
        name = "Parameters"

    timing: CeVIOParameter | None = field(
        default=None,
        metadata={
            "name": "Timing",
            "type": "Element",
        },
    )
    log_f0: CeVIOParameter | None = field(
        default=None,
        metadata={
            "name": "LogF0",
            "type": "Element",
        },
    )
    c0: CeVIOParameter | None = field(
        default=None,
        metadata={
            "name": "C0",
            "type": "Element",
        },
    )
    vib_amp: CeVIOParameter | None = field(
        default=None,
        metadata={
            "name": "VibAmp",
            "type": "Element",
        },
    )
    vib_frq: CeVIOParameter | None = field(
        default=None,
        metadata={
            "name": "VibFrq",
            "type": "Element",
        },
    )
    alpha: CeVIOParameter | None = field(
        default=None,
        metadata={
            "name": "Alpha",
            "type": "Element",
        },
    )
    husky: CeVIOParameter | None = field(
        default=None,
        metadata={
            "name": "Husky",
            "type": "Element",
        },
    )


class CeVIOSvss(BaseModel):
    class Meta:
        name = "SVSS"

    version: str | None = field(
        default="3.0.5",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )
    dictionary: CeVIODictionary | None = field(
        default_factory=CeVIODictionary,
        metadata={
            "name": "Dictionary",
            "type": "Element",
        },
    )
    sound_sources: CeVIOSoundSources = field(
        default_factory=CeVIOSoundSources,
        metadata={
            "name": "SoundSources",
            "type": "Element",
        },
    )


class CeVIOGeneration(BaseModel):
    class Meta:
        name = "Generation"

    author: CeVIOAuthor | None = field(
        default_factory=CeVIOAuthor,
        metadata={
            "name": "Author",
            "type": "Element",
        },
    )
    tts: CeVIOTts | None = field(
        default_factory=CeVIOTts,
        metadata={
            "name": "TTS",
            "type": "Element",
        },
    )
    svss: CeVIOSvss = field(
        default_factory=CeVIOSvss,
        metadata={
            "name": "SVSS",
            "type": "Element",
        },
    )


class CeVIOSong(BaseModel):
    class Meta:
        name = "Song"

    version: str | None = field(
        default="1.02",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )
    common_keys: bool | None = field(
        default=None,
        metadata={
            "name": "CommonKeys",
            "type": "Attribute",
        },
    )
    tempo: CeVIOTempo | None = field(
        default=None,
        metadata={
            "name": "Tempo",
            "type": "Element",
        },
    )
    beat: CeVIOBeat | None = field(
        default=None,
        metadata={
            "name": "Beat",
            "type": "Element",
        },
    )
    score: CeVIOScore = field(
        default_factory=CeVIOScore,
        metadata={
            "name": "Score",
            "type": "Element",
        },
    )
    parameter: CeVIOParameters = field(
        default_factory=CeVIOParameters,
        metadata={
            "name": "Parameter",
            "type": "Element",
        },
    )


class CeVIOBaseUnit(BaseModel):
    version: str | None = field(
        default="1.0",
        metadata={
            "name": "Version",
            "type": "Attribute",
        },
    )
    unit_id: str | None = field(
        default="",
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )
    group: str | None = field(
        default=None,
        metadata={
            "name": "Group",
            "type": "Attribute",
        },
    )
    start_time: XmlTime | None = field(
        default=None,
        metadata={
            "name": "StartTime",
            "type": "Attribute",
        },
    )
    duration: XmlTime | None = field(
        default=None,
        metadata={
            "name": "Duration",
            "type": "Attribute",
        },
    )
    category: str = field(
        default="SingerSong",
        metadata={
            "name": "Category",
            "type": "Attribute",
        },
    )
    text: str | None = field(
        default=None,
        metadata={
            "name": "Text",
            "type": "Attribute",
        },
    )
    snap_shot: str | None = field(
        default=None,
        metadata={
            "name": "SnapShot",
            "type": "Attribute",
        },
    )


class CeVIOSongUnit(CeVIOBaseUnit):
    song: CeVIOSong = field(
        default_factory=CeVIOSong,
        metadata={
            "name": "Song",
            "type": "Element",
        },
    )
    cast_id: str | None = field(
        default=None,
        metadata={
            "name": "CastId",
            "type": "Attribute",
        },
    )
    language: str | None = field(
        default="Japanese",
        metadata={
            "name": "Language",
            "type": "Attribute",
        },
    )


class CeVIOComponent(BaseModel):
    class Meta:
        name = "Component"

    name: str | None = field(
        default=None,
        metadata={
            "name": "Name",
            "type": "Attribute",
            "required": True,
        },
    )
    value: Decimal | None = field(
        default=None,
        metadata={
            "name": "Value",
            "type": "Attribute",
            "required": True,
        },
    )


class CeVIOPhoneme(BaseModel):
    class Meta:
        name = "Phoneme"

    data: str | None = field(
        default=None,
        metadata={
            "name": "Data",
            "type": "Attribute",
            "required": True,
        },
    )
    volume: Decimal | float | None = field(
        default=None,
        metadata={
            "name": "Volume",
            "type": "Attribute",
        },
    )
    speed: float | Decimal | None = field(
        default=None,
        metadata={
            "name": "Speed",
            "type": "Attribute",
        },
    )
    tone: Decimal | float | None = field(
        default=None,
        metadata={
            "name": "Tone",
            "type": "Attribute",
        },
    )


class CeVIODirection(BaseModel):
    class Meta:
        name = "Direction"

    volume: float | None = field(
        default=None,
        metadata={
            "name": "Volume",
            "type": "Attribute",
            "required": True,
        },
    )
    speed: float | None = field(
        default=None,
        metadata={
            "name": "Speed",
            "type": "Attribute",
            "required": True,
        },
    )
    tone: float | None = field(
        default=None,
        metadata={
            "name": "Tone",
            "type": "Attribute",
            "required": True,
        },
    )
    alpha: float | None = field(
        default=None,
        metadata={
            "name": "Alpha",
            "type": "Attribute",
            "required": True,
        },
    )
    log_f0_scale: float | None = field(
        default=None,
        metadata={
            "name": "LogF0Scale",
            "type": "Attribute",
            "required": True,
        },
    )
    component: list[CeVIOComponent] = field(
        default_factory=list,
        metadata={
            "name": "Component",
            "type": "Element",
            "min_occurs": 1,
        },
    )


class CeVIOPhonemes(BaseModel):
    class Meta:
        name = "Phonemes"

    phoneme: list[CeVIOPhoneme] = field(
        default_factory=list,
        metadata={
            "name": "Phoneme",
            "type": "Element",
            "min_occurs": 1,
        },
    )


class Word(BaseModel):
    class Meta:
        name = "word"

    phoneme: str | None = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    pronunciation: str | None = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    pos: str | None = field(
        default=None,
        metadata={
            "type": "Attribute",
        },
    )
    value: str = field(default="")


class AcousticPhrase(BaseModel):
    class Meta:
        name = "acoustic_phrase"

    word: list[Word] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )


class Base(BaseModel):
    acoustic_phrase: list[AcousticPhrase] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )


class Edited(BaseModel):
    acoustic_phrase: list[AcousticPhrase] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )


class CeVIOMetadataEN(BaseModel):
    class Meta:
        name = "Metadata_EN"

    base: Base | None = field(
        default=None,
        metadata={
            "name": "Base",
            "type": "Element",
            "required": True,
        },
    )
    edited: Edited | None = field(
        default=None,
        metadata={
            "name": "Edited",
            "type": "Element",
        },
    )


class CeVIOTalkUnit(CeVIOBaseUnit):
    metadata: str | None = field(
        default=None,
        metadata={
            "name": "Metadata",
            "type": "Element",
        },
    )
    metadata_en: CeVIOMetadataEN | None = field(
        default=None,
        metadata={
            "name": "Metadata_EN",
            "type": "Element",
        },
    )
    direction: CeVIODirection | None = field(
        default=None,
        metadata={
            "name": "Direction",
            "type": "Element",
            "required": True,
        },
    )
    phonemes: CeVIOPhonemes | None = field(
        default=None,
        metadata={
            "name": "Phonemes",
            "type": "Element",
            "required": True,
        },
    )


class CeVIOAudioUnit(CeVIOBaseUnit):
    file_path: str | None = field(
        default=None,
        metadata={
            "name": "FilePath",
            "type": "Attribute",
        },
    )


class CeVIOUnit(CeVIOSongUnit, CeVIOTalkUnit, CeVIOAudioUnit):
    class Meta:
        name = "Unit"


class CeVIOUnits(BaseModel):
    class Meta:
        name = "Units"

    unit: list[CeVIOUnit] = field(
        default_factory=list,
        metadata={
            "name": "Unit",
            "type": "Element",
        },
    )


class CeVIOScene(BaseModel):
    class Meta:
        name = "Scene"

    scene_id: str | None = field(
        default="",
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )
    timeline: CeVIOTimeline | None = field(
        default=None,
        metadata={
            "name": "Timeline",
            "type": "Element",
        },
    )
    talk_editor: CeVIOTalkEditor | None = field(
        default=None,
        metadata={
            "name": "TalkEditor",
            "type": "Element",
        },
    )
    song_editor: CeVIOSongEditor | None = field(
        default=None,
        metadata={
            "name": "SongEditor",
            "type": "Element",
        },
    )
    units: CeVIOUnits = field(
        default_factory=CeVIOUnits,
        metadata={
            "name": "Units",
            "type": "Element",
        },
    )
    groups: CeVIOGroups = field(
        default_factory=CeVIOGroups,
        metadata={
            "name": "Groups",
            "type": "Element",
        },
    )
    sound_setting: CeVIOSoundSetting | None = field(
        default_factory=CeVIOSoundSetting,
        metadata={
            "name": "SoundSetting",
            "type": "Element",
        },
    )


class CeVIOSequence(BaseModel):
    class Meta:
        name = "Sequence"

    sequence_id: str | None = field(
        default="",
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )
    scene: CeVIOScene = field(
        default_factory=CeVIOScene,
        metadata={
            "name": "Scene",
            "type": "Element",
        },
    )


class CeVIOCreativeStudioProject(BaseModel):
    class Meta:
        name = "Scenario"

    code: str | None = field(
        default="7251BC4B6168E7B2992FA620BD3E1E77",
        metadata={
            "name": "Code",
            "type": "Attribute",
        },
    )
    generation: CeVIOGeneration = field(
        default_factory=CeVIOGeneration,
        metadata={
            "name": "Generation",
            "type": "Element",
        },
    )
    sequence: CeVIOSequence = field(
        default_factory=CeVIOSequence,
        metadata={
            "name": "Sequence",
            "type": "Element",
        },
    )



import io
import pathlib

from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.serializers.config import SerializerConfig
from xsdata_pydantic.bindings import XmlParser, XmlSerializer

from libresvip.core.compat import ZipFile
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.utils.xmlutils import DefaultXmlWriter

from .models.container import Container
from .models.mxml4 import ScorePartwise
from .musicxml_generator import MusicXMLGenerator
from .musicxml_parser import MusicXMLParser
from .options import InputOptions, OutputOptions


class MusicXMLWriter(DefaultXmlWriter):
    def start_document(self) -> None:
        super().start_document()
        if self.config.xml_declaration:
            self.output.write(
                '<!DOCTYPE score-partwise PUBLIC "-//Recordare//DTD MusicXML 4.0 Partwise//EN" "http://www.musicxml.org/dtds/partwise.dtd">\n'
            )


class MusicXMLConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        content = path.read_bytes()
        xml_parser = XmlParser(config=ParserConfig(fail_on_unknown_properties=False))
        if content[:2] == b"PK":  # TODO: support mxl file extension
            with ZipFile(io.BytesIO(content), "r") as zf:
                container_content = zf.read("META-INF/container.xml")
                container = xml_parser.from_bytes(container_content, Container)
                first_file = zf.read(container.rootfiles.rootfile[0].full_path)
                score = xml_parser.from_bytes(first_file, ScorePartwise)
        else:
            score = xml_parser.from_bytes(content, ScorePartwise)
        return MusicXMLParser(options).parse_project(score)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        score = MusicXMLGenerator(options).generate_project(project)
        xml_serializer = XmlSerializer(
            config=SerializerConfig(pretty_print=True), writer=MusicXMLWriter
        )
        path.write_bytes(xml_serializer.render(score).encode("utf-8"))



import dataclasses

from libresvip.core.tick_counter import shift_tempo_list, skip_beat_list
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.search import find_last_index

from .deepvocal_pitch import (
    DvNoteWithPitch,
    DvSegmentPitchRawData,
    convert_note_key,
    pitch_from_dv_track,
)
from .model import (
    DvAudioTrack,
    DvNote,
    DvProject,
    DvSingingTrack,
    DvTempo,
    DvTimeSignature,
    DvTrackType,
)
from .options import InputOptions


@dataclasses.dataclass
class DeepVocalParser:
    options: InputOptions
    tick_prefix: int = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def parse_project(self, dv_project: DvProject) -> Project:
        time_signatures = self.parse_time_signatures(dv_project.inner_project.time_signatures)
        self.first_bar_length = round(time_signatures[0].bar_length())
        tempos = self.parse_tempos(dv_project.inner_project.tempos)
        instrumental_tracks = self.parse_instrumental_tracks(
            [
                track.track_data
                for track in dv_project.inner_project.tracks
                if track.track_type == DvTrackType.AUDIO and len(track.track_data.infos) > 0
            ]
        )
        singing_tracks = self.parse_singing_tracks(
            [
                track.track_data
                for track in dv_project.inner_project.tracks
                if track.track_type == DvTrackType.SINGING
            ],
            tempos,
        )
        return Project(
            time_signature_list=time_signatures,
            song_tempo_list=tempos,
            track_list=singing_tracks + instrumental_tracks,
        )

    def parse_time_signatures(
        self, dv_time_signatures: list[DvTimeSignature]
    ) -> list[TimeSignature]:
        self.tick_prefix = 0
        time_signatures = [
            TimeSignature(
                bar_index=time_signature.measure_position,
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
            )
            for time_signature in dv_time_signatures
        ]
        index = max(
            find_last_index(time_signatures, lambda beat: beat.bar_index <= 1),
            0,
        )
        for i in range(index + 1):
            if i < index:
                self.tick_prefix += (
                    time_signatures[i + 1].bar_index - time_signatures[i].bar_index
                ) * round(time_signatures[i].bar_length())
            else:
                self.tick_prefix += (1 - time_signatures[i].bar_index) * round(
                    time_signatures[i].bar_length()
                )
        return skip_beat_list(time_signatures, 1)

    def parse_tempos(self, dv_tempos: list[DvTempo]) -> list[SongTempo]:
        return [
            song_tempo if i > 0 else song_tempo.model_copy(update={"position": 0})
            for i, song_tempo in enumerate(
                shift_tempo_list(
                    [
                        SongTempo(
                            position=tempo.position,
                            bpm=tempo.bpm / 100,
                        )
                        for tempo in dv_tempos
                    ],
                    -self.tick_prefix + self.first_bar_length,
                )
            )
            if not i or song_tempo.position >= 0
        ]

    def parse_instrumental_tracks(
        self, dv_audio_tracks: list[DvAudioTrack]
    ) -> list[InstrumentalTrack]:
        track_list = []
        if self.options.import_instrumental_track:
            for dv_track in dv_audio_tracks:
                track = InstrumentalTrack(
                    title=dv_track.name or dv_track.infos[0].name,
                    mute=dv_track.mute,
                    solo=dv_track.solo,
                    offset=dv_track.infos[0].start + self.tick_prefix,
                    audio_file_path=dv_track.infos[0].path,
                )
                track_list.append(track)
        return track_list

    def parse_singing_tracks(
        self,
        dv_singing_tracks: list[DvSingingTrack],
        tempo_list: list[SongTempo],
    ) -> list[SingingTrack]:
        track_list = []
        for dv_track in dv_singing_tracks:
            for i, segment in enumerate(dv_track.segments):
                note_with_pitch: list[DvNoteWithPitch] = []
                tick_offset = segment.start
                track = SingingTrack(
                    title=f"{dv_track.name} {i + 1}",
                    mute=dv_track.mute,
                    solo=dv_track.solo,
                    ai_singer_name=segment.singer_name,
                    note_list=self.parse_notes(
                        segment.notes,
                        note_with_pitch,
                        tick_offset - self.tick_prefix,
                    ),
                )
                if (
                    self.options.import_pitch
                    and (
                        pitch := pitch_from_dv_track(
                            self.first_bar_length,
                            [
                                DvSegmentPitchRawData(
                                    tick_offset - self.tick_prefix,
                                    segment.pitch_data,
                                )
                            ],
                            note_with_pitch,
                            tempo_list,
                        )
                    )
                    is not None
                ):
                    track.edited_params.pitch = pitch
                track_list.append(track)
        return track_list

    def parse_notes(
        self,
        dv_notes: list[DvNote],
        note_with_pitch: list[DvNoteWithPitch],
        tick_offset: int,
    ) -> list[Note]:
        note_list = []
        for dv_note in dv_notes:
            note = Note(
                start_pos=tick_offset + dv_note.start,
                length=dv_note.length,
                key_number=convert_note_key(dv_note.key),
                lyric=dv_note.word,
                pronunciation=dv_note.phoneme if dv_note.phoneme != "-" else None,
            )
            note_with_pitch.append(
                DvNoteWithPitch(
                    note=note,
                    ben_dep=dv_note.ben_depth,
                    ben_len=dv_note.ben_length,
                    por_head=dv_note.por_head,
                    por_tail=dv_note.por_tail,
                    vibrato=dv_note.note_vibrato_data.vibrato_points,
                )
            )
            note_list.append(note)
        return note_list



import contextlib
import contextvars
import gettext
import sys

singleton_translation: gettext.NullTranslations | None = None
lazy_translation: contextvars.ContextVar[gettext.NullTranslations | None] = contextvars.ContextVar(
    "translator"
)


def gettext_lazy(message: str) -> str:
    if not message:
        return message
    with contextlib.suppress(LookupError):
        if (translation := singleton_translation) is not None or (
            translation := lazy_translation.get(None)
        ) is not None:
            return translation.gettext(message)
    return gettext.gettext(message)


def pgettext_lazy(context: str | None, message: str) -> str:
    if context is None:
        frame = sys._getframe(1)
        context = frame.f_globals.get("__package__")
    if context is None:
        return gettext_lazy(message)
    if not message:
        return message
    with contextlib.suppress(LookupError):
        if (translation := singleton_translation) is not None or (
            translation := lazy_translation.get(None)
        ) is not None:
            return translation.pgettext(context, message)
    return gettext.pgettext(context, message)



import io
import pathlib

from libresvip.core.compat import ZipFile, json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import VocaloidProject
from .options import InputOptions, OutputOptions
from .vpr_generator import VocaloidGenerator
from .vpr_parser import VocaloidParser


class VocaloidConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        with ZipFile(io.BytesIO(path.read_bytes()), "r") as archive_file:
            proj = VocaloidProject.model_validate_json(
                archive_file.read("Project/sequence.json"),
                context={
                    "extract_audio": options.extract_audio,
                    "path": path,
                    "archive_file": archive_file,
                },
            )
            return VocaloidParser(options, path).parse_project(proj)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        buffer = io.BytesIO()
        generator = VocaloidGenerator(options)
        vocaloid_project = generator.generate_project(project)
        with ZipFile(buffer, "w") as archive_file:
            archive_file.writestr(
                "Project/sequence.json",
                json.dumps(
                    vocaloid_project.model_dump(mode="json", exclude_none=True, by_alias=True),
                    ensure_ascii=False,
                ),
            )
            for wav_name, wav_path in generator.wav_paths.items():
                archive_file.writestr(f"Project/Audio/{wav_name}", wav_path.read_bytes())
        path.write_bytes(buffer.getvalue())



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from PySide6.QtCore import (
    QAbstractListModel,
    QByteArray,
    QModelIndex,
    QObject,
    QPersistentModelIndex,
    QStringListModel,
    Slot,
)

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

from libresvip.core.config import settings

if TYPE_CHECKING:
    from collections.abc import Iterator

    Item = dict[str, Any]
    Items = list[Item]
    Role2Name = dict[int, str]
    Name2Role = dict[str, int]
    RoleNames = dict[str, Any] | tuple[str, ...] | list[str]  # suggested


class ModelProxy(QAbstractListModel):
    """
    references:
        https://github.com/likianta/qmlease/blob/master/qmlease/qmlside/model/model.py
        https://pyblish.gitbooks.io/developer-guide/content/qml_and_python_interoperability.html
        https://stackoverflow.com/questions/54687953/declaring-a-qabstractlistmodel-as-a-property-in-pyside2
    """

    _defaults: dict[str, Any]  # see also `self._auto_complete`
    _items: Items
    _name_2_role: Name2Role
    _role_2_name: Role2Name

    def __init__(self, role_names: RoleNames) -> None:
        super().__init__(None)
        self._name_2_role = {x: i for i, x in enumerate(role_names)}
        self._role_2_name = dict(enumerate(role_names))
        self._items = []
        if isinstance(role_names, dict):
            self._defaults = role_names
        else:
            # trick: set default value to '' instead of None to improve
            # compatibility.
            self._defaults = dict.fromkeys(role_names, "")

    @property
    def _role_names(self) -> tuple[str, ...]:
        return tuple(self._name_2_role.keys())

    @property
    def items(self) -> Items:
        return self._items

    def __len__(self) -> int:
        return len(self._items)

    def __bool__(self) -> bool:
        return bool(self._items)

    def __getitem__(self, index: int) -> Item:
        return self._items[index]

    def __iter__(self) -> Iterator[Item]:
        return iter(self._items)

    # -------------------------------------------------------------------------
    # api
    # tip: all params which named `item` or `items` accept partial dict.

    @Slot(dict)
    def append(self, item: Item) -> None:
        model_index = QModelIndex()
        self.begin_insert_rows(
            model_index, self.row_count(model_index), self.row_count(model_index)
        )
        self._items.append(self._auto_complete(item))
        self.end_insert_rows()

    @Slot(list)
    def append_many(self, items: Items) -> None:
        model_index = QModelIndex()
        self.begin_insert_rows(
            model_index, self.row_count(model_index), self.row_count(model_index) + len(items) - 1
        )
        self._items.extend(map(self._auto_complete, items))
        self.end_insert_rows()

    @Slot(int, dict)
    def insert(self, index: int, item: Item) -> None:
        self.begin_insert_rows(QModelIndex(), index, index)
        self._items.insert(index, self._auto_complete(item))
        self.end_insert_rows()

    @Slot(int, list)
    def insert_many(self, index: int, items: Items) -> None:
        self.begin_insert_rows(QModelIndex(), index, index + len(items) - 1)
        self._items[index:index] = list(map(self._auto_complete, items))
        self.end_insert_rows()

    @Slot(result="QVariant")
    def pop(self) -> Item:
        self.begin_remove_rows(QModelIndex(), len(self._items) - 1, len(self._items) - 1)
        out = self._items.pop(0)
        self.end_remove_rows()
        return out

    @Slot(int, result="QVariant")
    def pop_many(self, count: int) -> Items:
        assert count > 0
        self.begin_remove_rows(QModelIndex(), len(self._items) - count, len(self._items) - 1)
        a, b = self._items[:-count], self._items[-count:]
        self._items = a
        self.end_remove_rows()
        return b

    @Slot(int)
    def delete(self, index: int) -> None:
        self.begin_remove_rows(QModelIndex(), index, index)
        self._items.pop(index)
        self.end_remove_rows()

    @Slot(int, int)
    def delete_many(self, index: int, count: int) -> None:
        assert count > 0
        self.begin_remove_rows(QModelIndex(), index, index + count - 1)
        a, _ = (
            self._items[:index] + self._items[index + count :],
            self._items[index : index + count],
        )
        self._items = a
        self.end_remove_rows()

    @Slot(int, int)
    def move(self, old_index: int, new_index: int) -> None:
        self.begin_move_rows(QModelIndex(), old_index, old_index, QModelIndex(), new_index)
        item = self._items.pop(old_index)
        self._items.insert(new_index, item)
        self.end_move_rows()

    @Slot(int, int, int)
    def move_many(self, old_index: int, new_index: int, count: int) -> None:
        assert count > 0
        self.begin_move_rows(
            QModelIndex(),
            old_index,
            old_index + count - 1,
            QModelIndex(),
            new_index,
        )
        items = self._items[old_index : old_index + count]
        self._items[old_index : old_index + count] = []
        self._items[new_index:new_index] = items
        self.end_move_rows()

    @Slot(int, result=bool)
    def move_up(self, index: int) -> bool:
        if index > 0:
            self.move(index, index - 1)
            return True
        return False

    @Slot(int, result=bool)
    def move_down(self, index: int) -> bool:
        if index < len(self._items) - 1:
            self.move(index, index + 1)
            return True
        return False

    @Slot()
    def clear(self) -> None:
        self.begin_remove_rows(QModelIndex(), 0, len(self._items) - 1)
        self._items.clear()
        self.end_remove_rows()

    @Slot(int, result="QVariant")
    def get(self, index: int) -> Item:
        return self._items[index]

    @Slot(int, result="QVariant")
    @Slot(int, int, result="QVariant")
    def get_many(
        self,
        start: int | None = None,
        end: int | None = None,  # TODO: shall we use `count` instead?
    ) -> Items:
        if start is not None and end is None:
            start, end = 0, start
        else:
            assert start is not None
            assert end is not None
        return self._items[start:end]

    @Slot(int, dict)
    def update(self, index: int, item: Item) -> None:
        if index >= len(self._items):
            return
        self._items[index].update(item)
        # emit signal of `self.dataChanged` to notify qml side that some item
        # has been changed.
        # `dataChanged.emit` accepts two arguments:
        #   dataChanged.emit(QModelIndex start, QModelIndex end)
        # how to create QModelIndex instance: use `self.createIndex(row, col)`.
        # ref: https://blog.csdn.net/LaoYuanPython/article/details/102011031
        qindex = self.create_index(index, 0)
        self.dataChanged.emit(qindex, qindex, [self._name_2_role[x] for x in item])

    @Slot(int, list)
    def update_many(self, start: int, items: Items) -> None:
        if not items:
            return
        end = start + len(items)
        for old, new in zip(self._items[start:end], items):
            old.update(new)
        qindex_start = self.create_index(start, 0)
        qindex_end = self.create_index(end - 1, 0)
        self.dataChanged.emit(qindex_start, qindex_end)

    def _auto_complete(self, item: Item) -> Item:
        for k, v in self._defaults.items():
            if k not in item:
                item[k] = v
        return item

    # -------------------------------------------------------------------------
    # overrides

    def data(self, index: QModelIndex | QPersistentModelIndex, /, role: int = ...) -> Any:  # type: ignore[assignment]
        name = self._role_2_name[role]
        return self._items[index.row()].get(name, "")

    def set_data(
        self,
        index: QModelIndex | QPersistentModelIndex,
        value: Any,
        /,
        role: int = ...,  # type: ignore[assignment]
    ) -> bool:
        # 获取指定角色的名称
        name = self._role_names[role]
        # 将指定角色的值设置为传入的值
        self._items[index.row()][name] = value
        # 发射数据改变信号
        self.dataChanged.emit(index, index)
        # 返回True表示设置成功
        return True

    def row_count(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:  # type: ignore[assignment]
        return len(self._items)

    def role_names(self) -> dict[int, QByteArray]:
        # return self._role_2_name
        return {k: QByteArray(v.encode("utf-8")) for k, v in self._role_2_name.items()}


class LyricReplacementPresetsModel(QStringListModel):
    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self.set_string_list(list(settings.lyric_replace_rules))

    @Slot(str)
    def append(self, value: str) -> None:
        settings.lyric_replace_rules.setdefault(value, [])
        self.set_string_list(list(settings.lyric_replace_rules))

    @Slot(str)
    def remove(self, value: str) -> None:
        settings.lyric_replace_rules.pop(value, None)
        self.set_string_list(list(settings.lyric_replace_rules))



import dataclasses
import math
import pathlib
from collections.abc import Callable

from libresvip.core.constants import DEFAULT_CHINESE_LYRIC
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import db_to_float

from .model import (
    VocalShifterLabel,
    VocalShifterNote,
    VocalShifterPatternData,
    VocalShifterPatternMetadata,
    VocalShifterPatternType,
    VocalShifterProjectData,
    VocalShifterProjectMetadata,
    VocalShifterTrackMetadata,
)
from .options import InputOptions
from .utils import ansi2unicode


@dataclasses.dataclass
class VocalShifterParser:
    options: InputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    tick_rate: float = dataclasses.field(init=False)

    def parse_project(self, vshp_proj: VocalShifterProjectData) -> Project:
        project = Project(
            song_tempo_list=self.parse_tempo(vshp_proj.project_metadata),
            time_signature_list=self.parse_time_signature(vshp_proj.project_metadata),
        )
        project.track_list = self.parse_track_list(vshp_proj)
        return project

    def parse_tempo(self, vshp_metadata: VocalShifterProjectMetadata) -> list[SongTempo]:
        tempo_list = [
            SongTempo(
                position=0,
                bpm=vshp_metadata.tempo,
            )
        ]
        self.synchronizer = TimeSynchronizer(tempo_list)
        self.tick_rate = vshp_metadata.tempo / 25
        return tempo_list

    def parse_time_signature(
        self, vshp_metadata: VocalShifterProjectMetadata
    ) -> list[TimeSignature]:
        return [
            TimeSignature(
                bar_index=0,
                numerator=vshp_metadata.numerator,
                denominator=vshp_metadata.denominator,
            )
        ]

    def parse_track_list(self, vshp_proj: VocalShifterProjectData) -> list[Track]:
        track_list = []
        for pattern_metadata, pattern_data in zip(
            vshp_proj.pattern_metadatas, vshp_proj.pattern_datas
        ):
            if (
                pattern_data.header.pattern_type == VocalShifterPatternType.WAVE
                and not self.options.wave_to_singing
            ):
                track = self.parse_instrumental_track(
                    pattern_metadata,
                    pattern_data,
                    vshp_proj.track_metadatas[pattern_metadata.track_index - 1],
                )
            else:
                track = self.parse_singing_track(
                    pattern_metadata,
                    pattern_data,
                    vshp_proj.track_metadatas[pattern_metadata.track_index - 1],
                )
            track_list.append(track)
        return track_list

    def parse_instrumental_track(
        self,
        pattern_metadata: VocalShifterPatternMetadata,
        pattern_data: VocalShifterPatternData,
        track_metadata: VocalShifterTrackMetadata,
    ) -> InstrumentalTrack:
        sample_rate = pattern_data.header.sample_rate
        sample_offset = pattern_metadata.offset_samples + pattern_metadata.offset_correction
        offset_in_seconds = sample_offset / sample_rate
        offset_in_ticks = self.synchronizer.get_actual_ticks_from_secs(offset_in_seconds)
        return InstrumentalTrack(
            audio_file_path=ansi2unicode(pattern_metadata.path_and_ext.split(b"\x00")[0]),
            offset=offset_in_ticks,
            solo=track_metadata.solo,
            mute=track_metadata.mute,
            volume=track_metadata.volume,
            pan=track_metadata.pan,
        )

    def parse_singing_track(
        self,
        pattern_metadata: VocalShifterPatternMetadata,
        pattern_data: VocalShifterPatternData,
        track_metadata: VocalShifterTrackMetadata,
    ) -> SingingTrack:
        file_path = pathlib.Path(ansi2unicode(pattern_metadata.path_and_ext.split(b"\x00")[0]))
        track = SingingTrack(
            title=file_path.stem,
            solo=track_metadata.solo,
            mute=track_metadata.mute,
            volume=track_metadata.volume,
            pan=track_metadata.pan,
        )
        sample_offset = pattern_metadata.offset_samples + pattern_metadata.offset_correction
        offset_in_seconds = sample_offset / pattern_data.header.sample_rate
        offset_in_ticks = int(self.synchronizer.get_actual_ticks_from_secs(offset_in_seconds))
        track.note_list = self.parse_note_list(
            offset_in_ticks,
            pattern_data.notes.notes,
            pattern_data.labels.labels if pattern_data.labels else [],
        )
        track.edited_params = self.parse_params(offset_in_seconds, pattern_data)
        return track

    def parse_note_list(
        self,
        offset: int,
        notes: list[VocalShifterNote],
        labels: list[VocalShifterLabel],
    ) -> list[Note]:
        note_list = []
        if labels:
            for label in labels:
                target_note = next(
                    (
                        note
                        for note in notes
                        if note.start_tick <= label.start_tick <= note.start_tick + note.length
                    ),
                    None,
                )
                note_list.append(
                    Note(
                        start_pos=offset + round(label.start_tick * self.tick_rate),
                        length=int((label.end_tick - label.start_tick) * self.tick_rate),
                        key_number=target_note.pitch // 100 if target_note is not None else 60,
                        lyric=ansi2unicode(label.name.partition(b"\x00")[0]),
                    )
                )
        else:
            note_list.extend(
                Note(
                    start_pos=offset + round(note.start_tick * self.tick_rate),
                    length=int(note.length * self.tick_rate),
                    key_number=note.pitch // 100,
                    lyric=DEFAULT_CHINESE_LYRIC,
                )
                for note in notes
            )
        return note_list

    @staticmethod
    def volume_to_value(volume: float) -> float:
        db = math.log(volume, 1 + 1 / 9)
        float_value = db_to_float(db)
        return round(1000 * float_value - 1000)

    def parse_params(
        self,
        offset: float,
        pattern_data: VocalShifterPatternData,
    ) -> Params:
        params = Params(
            pitch=self.parse_pitch_curve(offset, pattern_data),
        )
        if self.options.import_dynamics:
            params.volume = self.parse_param_curve(
                offset,
                pattern_data,
                "dyn" if self.options.use_edited_dynamics else "ori_dyn",
                self.volume_to_value,
            )
        if self.options.import_breath:
            params.breath = self.parse_param_curve(
                offset,
                pattern_data,
                "bre",
                (10.0).__rfloordiv__,
            )
        if self.options.import_formant:
            params.gender = self.parse_param_curve(
                offset,
                pattern_data,
                "frm",
                (10.0).__rfloordiv__,
            )
        return params

    def parse_pitch_curve(
        self,
        offset: float,
        pattern_data: VocalShifterPatternData,
    ) -> ParamCurve:
        pitch_curve = ParamCurve()
        pitch_curve.points.append(Point.start_point())
        has_pitch = False
        time_step = 1 / pattern_data.header.points_per_second
        for point in pattern_data.points:
            value = point.pit if self.options.use_edited_pitch else point.ori_pit
            if value != 0:
                if not has_pitch:
                    pitch_curve.points.append(
                        Point(
                            round(self.synchronizer.get_actual_ticks_from_secs(offset)) + 1920,
                            -100,
                        )
                    )
                    has_pitch = True
                pitch_curve.points.append(
                    Point(
                        round(self.synchronizer.get_actual_ticks_from_secs(offset)) + 1920,
                        value,
                    )
                )
            elif has_pitch:
                pitch_curve.points.append(
                    Point(
                        round(self.synchronizer.get_actual_ticks_from_secs(offset)) + 1920,
                        -100,
                    )
                )
                has_pitch = False
            offset += time_step
        pitch_curve.points.append(Point.end_point())
        return pitch_curve

    def parse_param_curve(
        self,
        offset: float,
        pattern_data: VocalShifterPatternData,
        attr_name: str,
        mapping_func: Callable[[int], float],
    ) -> ParamCurve:
        param_curve = ParamCurve()
        time_step = 1 / pattern_data.header.points_per_second
        for point in pattern_data.points:
            value = getattr(point, attr_name)
            param_curve.points.append(
                Point(
                    round(self.synchronizer.get_actual_ticks_from_secs(offset)) + 1920,
                    int(mapping_func(value)),
                )
            )
            offset += time_step
        return param_curve



import dataclasses

import pypinyin

from libresvip.model.base import (
    Note,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point

from .model import Y77Note, Y77Project
from .options import InputOptions


@dataclasses.dataclass
class Y77Parser:
    options: InputOptions
    first_bar_length: int = dataclasses.field(init=False)

    def parse_project(self, y77_project: Y77Project) -> Project:
        time_signatures = self.parse_time_signatures(y77_project)
        self.first_bar_length = round(time_signatures[0].bar_length())
        return Project(
            song_tempo_list=self.parse_tempos(y77_project.bpm),
            time_signature_list=time_signatures,
            track_list=self.parse_tracks(y77_project.notes),
        )

    def parse_tempos(self, bpm: float) -> list[SongTempo]:
        return [SongTempo(bpm=bpm, position=0)]

    def parse_time_signatures(self, y77_project: Y77Project) -> list[TimeSignature]:
        return [
            TimeSignature(
                numerator=y77_project.bbar,
                denominator=y77_project.bbeat,
            )
        ]

    def parse_tracks(self, notes: list[Y77Note]) -> list[SingingTrack]:
        return [
            SingingTrack(
                ai_singer_name="元七七",
                note_list=self.parse_notes(notes),
                edited_params=self.parse_params(notes),
            )
        ]

    def parse_notes(self, notes: list[Y77Note]) -> list[Note]:
        note_list = []
        for y77_note in notes:
            note = Note(
                lyric=y77_note.lyric,
                start_pos=y77_note.start * 30,
                length=y77_note.length * 30,
                key_number=88 - y77_note.pitch,
            )
            phonemes = pypinyin.pinyin(y77_note.lyric, heteronym=True, style=pypinyin.STYLE_NORMAL)
            if len(phonemes[0]) > 1 or phonemes[0][0] != y77_note.py:
                note.pronunciation = y77_note.py
            note_list.append(note)
        return note_list

    def parse_params(self, notes: list[Y77Note]) -> Params:
        params = Params()
        for y77_note in notes:
            if self.options.import_pitch and len(y77_note.pit):
                step = y77_note.length * 30 / (len(y77_note.pit) - 1)
                pbs = y77_note.pbs + 1
                params.pitch.points.append(
                    Point(y77_note.start * 30 - 5 + self.first_bar_length, -100)
                )
                for i in range(len(y77_note.pit)):
                    params.pitch.points.append(
                        Point(
                            round(y77_note.start * 30 + i * step) + self.first_bar_length,
                            round(((y77_note.pit[i] - 50) / 50 * pbs + 88 - y77_note.pitch) * 100),
                        )
                    )
                params.pitch.points.append(
                    Point(
                        (y77_note.start + y77_note.length) * 30 + 5 + self.first_bar_length,
                        -100,
                    )
                )
        return params



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.utils.text import to_unicode

from .model import TuneLabProject
from .options import InputOptions, OutputOptions
from .tunelab_generator import TuneLabGenerator
from .tunelab_parser import TuneLabParser


class TuneLabConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        tlp_project = TuneLabProject.model_validate_json(to_unicode(path.read_bytes()))
        return TuneLabParser(options).parse_project(tlp_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        tlp_project = TuneLabGenerator(options).generate_project(project)
        path.write_bytes(
            json.dumps(
                tlp_project.model_dump(mode="json", by_alias=True),
                ensure_ascii=False,
            ).encode("utf-8")
        )



import sys

try:
    import ujson as json
except ImportError:
    import json

    if hasattr(json, "_default_encoder"):
        json._default_encoder.item_separator = ","
        json._default_encoder.key_separator = ":"

try:
    import pyzstd as zstd
except ImportError:
    try:
        import numcodecs.zstd as zstd
    except ImportError:
        import zstandard as zstd

__all__ = ["Traversable", "ZipFile", "json", "zstd"]

if sys.version_info < (3, 11):
    from importlib_resources.abc import Traversable
    from repro_zipfile import ReproducibleZipFile as ZipFile
else:
    from importlib.resources.abc import Traversable
    from zipfile import ZipFile



# mypy: disable-error-code="attr-defined"
from __future__ import annotations

import enum
from types import GenericAlias
from typing import Annotated, Any, Literal, get_args

from packaging.version import Version
from pydantic import AliasChoices, Field, ValidationInfo, field_validator

from libresvip.core.exceptions import UnsupportedProjectVersionError
from libresvip.model.base import BaseModel
from libresvip.utils.translation import gettext_lazy as _

from .value_tree import JUCENode, JUCEVarTypes


class VoiSonaPlayControlItem(BaseModel):
    loop: bool = Field(alias="Loop")
    loop_start: float = Field(0, alias="LoopStart")
    loop_end: float = Field(alias="LoopEnd")
    play_position: float = Field(alias="PlayPosition")


class VoiSonaSongEditorItem(BaseModel):
    editor_width: int = Field(1920, alias="EditorWidth")
    editor_height: int = Field(1080, alias="EditorHeight")


class VoiSonaControlPanelStatus(BaseModel):
    quantization: int = Field(alias="Quantization")
    edit_tool: int | None = Field(None, alias="EditTool")
    record_note: bool = Field(True, alias="RecordNote")
    record_tempo: bool = Field(True, alias="RecordTempo")


class VoiSonaAdjustToolBarStatus(BaseModel):
    main_panel: int = Field(alias="MainPanel")
    sub_panel: int = Field(alias="SubPanel")


class VoiSonaPanelControllerStatus(BaseModel):
    tempo_panel: bool | None = Field(None, alias="TempoPanel")
    beat_panel: bool | None = Field(None, alias="BeatPanel")
    key_panel: bool | None = Field(None, alias="KeyPanel")
    dynamics_panel: bool | None = Field(None, alias="DynamicsPanel")
    global_param_panel: bool | None = Field(None, alias="GlobalParamPanel")
    lyric_panel: bool | None = Field(None, alias="LyricPanel")
    property_panel: bool | None = Field(None, alias="PropertyPanel")


class VoiSonaMainPanelStatus(VoiSonaPanelControllerStatus):
    scale_x: float = Field(alias="ScaleX_V2", validation_alias=AliasChoices("ScaleX_V2", "ScaleX"))
    scale_y: float = Field(alias="ScaleY_V2", validation_alias=AliasChoices("ScaleY_V2", "ScaleY"))
    scroll_x: float | None = Field(None, alias="ScrollX")
    scroll_y: float | None = Field(None, alias="ScrollY")


class VoiSonaNeuralVocoderInformation(BaseModel):
    nv_file_name: str = Field(alias="NVFileName")
    nv_lib_file_name: str | None = Field(None, alias="NVLibFileName")
    nv_version: str = Field(alias="NVVersion")
    nv_hash: str | None = Field(None, alias="NVHash")
    nv_lib_hash: str | None = Field(None, alias="NVLibHash")
    nv_label: str = Field(alias="NVLabel")


class VoiSonaNeuralVocoderListItem(BaseModel):
    neural_vocoder_information: list[VoiSonaNeuralVocoderInformation] | None = Field(
        None, alias="NeuralVocoderInformation"
    )


class VoiSonaEmotionItem(BaseModel):
    label: str = Field(alias="Label")
    ratio: float = Field(alias="Ratio")


class VoiSonaEmotionListItem(BaseModel):
    emotion: list[VoiSonaEmotionItem] = Field(default_factory=list, alias="Emotion")


class VoiSonaSpecialSymbol(BaseModel):
    special_symbol_raspy: list[str] = Field(default_factory=list, alias="SpecialSymbolRaspy")
    special_symbol_falsetto: list[str] = Field(default_factory=list, alias="SpecialSymbolFalsetto")


class VoiSonaVoiceInformation(BaseModel):
    neural_vocoder_list: list[VoiSonaNeuralVocoderListItem] = Field(
        default_factory=list, alias="NeuralVocoderList"
    )
    emotion_list: list[VoiSonaEmotionListItem] = Field(default_factory=list, alias="EmotionList")
    character_name: str = Field(alias="CharacterName")
    language: str = Field("ja_JP", alias="Language")
    active_after_this_version: str | None = Field(None, alias="ActiveAfterThisVersion")
    voice_file_name: str = Field(alias="VoiceFileName")
    voice_lib_file_name: str | None = Field(None, alias="VoiceLibFileName")
    voice_version: str = Field(alias="VoiceVersion")
    voice_hash: str | None = Field(None, alias="VoiceHash")
    voice_lib_hash: str | None = Field(None, alias="VoiceLibHash")
    special_symbol: list[VoiSonaSpecialSymbol] = Field(default_factory=list, alias="SpecialSymbol")


class VoiSonaGlobalParameter(BaseModel):
    global_vib_amp: float = Field(1, alias="GlobalVibAmp")
    global_vib_frq: float = Field(0, alias="GlobalVibFrq")
    global_alpha: float = Field(0, alias="GlobalAlpha")
    global_husky: float = Field(0, alias="GlobalHusky")
    global_tune: float | None = Field(None, alias="GlobalTune")


class VoiSonaSoundItem(BaseModel):
    clock: int = Field(alias="Clock")
    tempo: float = Field(alias="Tempo")


class VoiSonaTempoItem(BaseModel):
    sound: list[VoiSonaSoundItem] = Field(default_factory=list, alias="Sound")


class VoiSonaTimeItem(BaseModel):
    clock: int = Field(alias="Clock")
    beats: int = Field(alias="Beats")
    beat_type: int = Field(alias="BeatType")


class VoiSonaBeatItem(BaseModel):
    time: list[VoiSonaTimeItem] = Field(default_factory=list, alias="Time")


class VoiSonaKeyItem(BaseModel):
    clock: int = Field(alias="Clock")
    fifths: int = Field(alias="Fifths")
    mode: int = Field(alias="Mode")


class VoiSonaDynamic(BaseModel):
    clock: int = Field(alias="Clock")
    value: int = Field(alias="Value")


class VoiSonaNoteItem(BaseModel):
    clock: int = Field(alias="Clock")
    duration: int = Field(alias="Duration")
    pitch_step: int = Field(alias="PitchStep")
    pitch_octave: int = Field(alias="PitchOctave")
    lyric: str = Field(alias="Lyric")
    syllabic: int = Field(alias="Syllabic")
    phoneme: str = Field(alias="Phoneme")
    do_re_mi: bool | None = Field(None, alias="DoReMi")
    accent: bool | None = Field(None, alias="Accent")
    breath: bool | None = Field(None, alias="Breath")
    staccato: bool | None = Field(None, alias="Staccato")
    slur_start: bool | None = Field(None, alias="SlurStart")
    slur_stop: bool | None = Field(None, alias="SlurStop")
    default_phoneme: str | None = Field(None, alias="DefaultPhoneme")
    past_analyzed_phoneme: str | None = Field(None, alias="PastAnalyzedPhoneme")
    special_symbol_raspy: bool | None = Field(None, alias="SpecialSymbolRaspy")
    special_symbol_falsetto: bool | None = Field(None, alias="SpecialSymbolFalsetto")


class VoiSonaScoreItem(BaseModel):
    key: list[VoiSonaKeyItem] = Field(default_factory=list, alias="Key")
    dynamics: list[VoiSonaDynamic] = Field(default_factory=list, alias="Dynamics")
    note: list[VoiSonaNoteItem] = Field(default_factory=list, alias="Note")


class VoiSonaSongItem(BaseModel):
    tempo: list[VoiSonaTempoItem] = Field(default_factory=list, alias="Tempo")
    beat: list[VoiSonaBeatItem] = Field(default_factory=list, alias="Beat")
    score: list[VoiSonaScoreItem] = Field(default_factory=list, alias="Score")


class VoiSonaPointData(BaseModel):
    index: int | None = Field(None, alias="Index")
    repeat: int | None = Field(None, alias="Repeat")
    value: float = Field(alias="Value")


class VoiSonaParameterItem(BaseModel):
    length: int = Field(alias="Length")
    data: list[VoiSonaPointData] = Field(default_factory=list, alias="Data")


class VoiSonaParametersItem(BaseModel):
    timing: list[VoiSonaParameterItem] | None = Field(None, alias="Timing")
    c0: list[VoiSonaParameterItem] | None = Field(None, alias="C0")
    c0_c_tick: list[VoiSonaParameterItem] | None = Field(None, alias="C0CTick")
    log_f0: list[VoiSonaParameterItem] | None = Field(None, alias="LogF0")
    log_f0_c_tick: list[VoiSonaParameterItem] | None = Field(None, alias="LogF0CTick")
    vocoder_log_f0: float | None = Field(None, alias="VocoderLogF0")
    vib_amp: list[VoiSonaParameterItem] | None = Field(None, alias="VibAmp")
    vib_amp_c_tick: list[VoiSonaParameterItem] | None = Field(None, alias="VibAmpCTick")
    vib_frq: list[VoiSonaParameterItem] | None = Field(None, alias="VibFrq")
    vib_frq_c_tick: list[VoiSonaParameterItem] | None = Field(None, alias="VibFrqCTick")
    alpha: list[VoiSonaParameterItem] | None = Field(None, alias="Alpha")
    alpha_c_tick: list[VoiSonaParameterItem] | None = Field(None, alias="AlphaCTick")
    husky: list[VoiSonaParameterItem] | None = Field(None, alias="Husky")
    husky_c_tick: list[VoiSonaParameterItem] | None = Field(None, alias="HuskyCTick")


class VoiSonaSignerConfig(BaseModel):
    snap_shot: str | None = Field(None, alias="SnapShot")


class VoiSonaStateInformation(BaseModel):
    song_editor: list[VoiSonaSongEditorItem] = Field(default_factory=list, alias="SongEditor")
    control_panel_status: list[VoiSonaControlPanelStatus] = Field(
        default_factory=list, alias="ControlPanelStatus"
    )
    adjust_tool_bar_status: list[VoiSonaAdjustToolBarStatus] = Field(
        default_factory=list, alias="AdjustToolBarStatus"
    )
    main_panel_status: list[VoiSonaMainPanelStatus] = Field(
        default_factory=list, alias="MainPanelStatus"
    )
    panel_controller_status: list[VoiSonaPanelControllerStatus] | None = Field(
        None, alias="PanelControllerStatus"
    )
    voice_information: list[VoiSonaVoiceInformation] | None = Field(None, alias="VoiceInformation")
    global_parameters: list[VoiSonaGlobalParameter] | None = Field(None, alias="GlobalParameters")
    song: list[VoiSonaSongItem] | None = Field(None, alias="Song")
    parameter: list[VoiSonaParametersItem] | None = Field(None, alias="Parameter")
    tempo_sync: bool | None = Field(False, alias="TempoSync")
    signer_config: list[VoiSonaSignerConfig] | None = Field(None, alias="SignerConfig")
    version_of_app_file_saved: str | None = Field("1.8.0.17", alias="VersionOfAppFileSaved")


class VoiSonaPluginData(BaseModel):
    state_information: VoiSonaStateInformation = Field(
        default_factory=VoiSonaStateInformation, alias="StateInformation"
    )


class VoiSonaAudioEventItem(BaseModel):
    path: str = Field(alias="Path")
    offset: float = Field(alias="Offset")


class VoiSonaTrackType(enum.IntEnum):
    SINGING = 0
    AUDIO = 2


class VoiSonaBaseTrackItem(BaseModel):
    name: str = Field(alias="Name")
    state: int = Field(0, alias="State")
    volume: float = Field(0, alias="Volume")
    pan: float = Field(0, alias="Pan")


class VoiSonaAudioTrackItem(VoiSonaBaseTrackItem):
    track_type: Literal[VoiSonaTrackType.AUDIO] = Field(VoiSonaTrackType.AUDIO, alias="Type")
    audio_event: list[VoiSonaAudioEventItem] | None = Field(None, alias="AudioEvent")


class VoiSonaSingingTrackItem(VoiSonaBaseTrackItem):
    track_type: Literal[VoiSonaTrackType.SINGING] = Field(VoiSonaTrackType.SINGING, alias="Type")
    plugin_data: VoiSonaPluginData = Field(default_factory=VoiSonaPluginData, alias="PluginData")


VoiSonaTrackItem = Annotated[
    VoiSonaAudioTrackItem | VoiSonaSingingTrackItem,
    Field(discriminator="track_type"),
]


class VoiSonaTrack(BaseModel):
    track: list[VoiSonaTrackItem] = Field(default_factory=list, alias="Track")


class VoiSonaGuiStatus(BaseModel):
    scale_x: float = Field(100, alias="ScaleX")
    scale_y: float = Field(100, alias="ScaleY")
    grid_mode: int | None = Field(None, alias="GridMode")
    grid_index: int | None = Field(None, alias="GridIndex")


class VoiSonaProject(BaseModel):
    play_control: list[VoiSonaPlayControlItem] = Field(default_factory=list, alias="PlayControl")
    tracks: list[VoiSonaTrack] = Field(default_factory=list, alias="Tracks")
    gui_status: list[VoiSonaGuiStatus] = Field(default_factory=list, alias="GUIStatus")
    version_of_app_file_saved: str | None = Field("1.8.0.17", alias="VersionOfAppFileSaved")

    @field_validator("version_of_app_file_saved")
    @classmethod
    def version_validator(cls, value: str, _info: ValidationInfo) -> str:
        if Version(value) < Version("1.8"):
            msg = _("Unsupported project version") + f" {value}"
            raise UnsupportedProjectVersionError(msg)
        return value


def value_to_dict(field_value: Any, field_type: type) -> dict[str, Any]:
    if issubclass(field_type, bool):
        variant_type = JUCEVarTypes.BOOL_TRUE if field_value is True else JUCEVarTypes.BOOL_FALSE
    elif issubclass(field_type, enum.IntEnum | int):
        variant_type = JUCEVarTypes.INT
    elif issubclass(field_type, float):
        variant_type = JUCEVarTypes.DOUBLE
    elif issubclass(field_type, str):
        variant_type = JUCEVarTypes.STRING
    elif issubclass(field_type, bytes):
        variant_type = JUCEVarTypes.BINARY
    elif issubclass(field_type, list):
        variant_type = JUCEVarTypes.ARRAY
        field_value = [value_to_dict(item, type(item)) for item in field_value]
    else:
        variant_type = None
        msg = f"Unknown field type {field_type}"
        raise TypeError(msg)
    return {
        "type": variant_type,
        "value": field_value,
    }


def model_to_value_tree(model: BaseModel, name: str = "TSSolution") -> dict[str, Any]:
    model_class = type(model)
    value_tree = {
        "name": name,
        "attrs": [],
        "children": [],
    }
    for field_name, field_info in model_class.model_fields.items():
        field_value = getattr(model, field_name)
        alias_field_name = field_info.alias or field_name
        if field_value is not None:
            if isinstance(field_info.annotation, type):
                field_type = field_info.annotation
            elif field_info.default is None or field_info.default_factory is list:
                field_type = field_info.annotation
                while not isinstance(field_type, type | GenericAlias):
                    field_type = get_args(field_type)[0]
            else:
                field_type = type(field_info.default)
            if isinstance(field_type, GenericAlias):
                inner_type = get_args(field_type)[0]
                if not isinstance(inner_type, type) or issubclass(inner_type, BaseModel):
                    value_tree["children"].extend(
                        model_to_value_tree(item, alias_field_name) for item in field_value
                    )
                else:
                    value_tree["attrs"].extend(
                        {"name": alias_field_name, "data": value_to_dict(item, inner_type)}
                        for item in field_value
                    )
            elif alias_field_name == "PluginData":
                value_tree["attrs"].append(
                    {
                        "name": alias_field_name,
                        "data": {
                            "type": JUCEVarTypes.BINARY,
                            "value": JUCENode.build(
                                model_to_value_tree(
                                    field_value.state_information,
                                    "StateInformation",
                                )
                            ),
                        },
                    }
                )
            elif issubclass(field_type, BaseModel):
                value_tree["children"].append(model_to_value_tree(field_value, alias_field_name))
            else:
                value_tree["attrs"].append(
                    {"name": alias_field_name, "data": value_to_dict(field_value, field_type)}
                )
    return value_tree



import math
from dataclasses import dataclass

from libresvip.core.constants import (
    DEFAULT_PITCH_BEND_SENSITIVITY,
    MAX_PITCH_BEND_SENSITIVITY,
    MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS,
    PITCH_MAX_VALUE,
)
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, ParamCurve
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.point import Point
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve
from libresvip.utils.music_math import clamp


@dataclass
class ControllerEvent:
    pos: int
    value: int


@dataclass
class VocaloidPartPitchData:
    start_pos: int
    pit: list[ControllerEvent]
    pbs: list[ControllerEvent]


def pitch_from_vocaloid_parts(
    data_by_parts: list[VocaloidPartPitchData],
    synchronizer: TimeSynchronizer,
    note_list: list[Note],
    vibrato_rate_interval_dict: PiecewiseIntervalDict,
    vibrato_depth_interval_dict: PiecewiseIntervalDict,
    first_bar_length: int,
    lower_bound: int,
    upper_bound: int,
) -> ParamCurve | None:
    pitch_raw_data_by_part = []
    for part in data_by_parts:
        pit = part.pit
        pbs = part.pbs
        pit_multiplied_by_pbs = {}
        pit_index = 0
        pbs_current_value = DEFAULT_PITCH_BEND_SENSITIVITY
        for pbs_event in pbs:
            for i in range(pit_index, len(pit)):
                pit_event = pit[i]
                if pit_event.pos < pbs_event.pos:
                    pit_multiplied_by_pbs[pit_event.pos + part.start_pos] = (
                        pit_event.value * pbs_current_value
                    )
                    if i == len(pit) - 1:
                        pit_index = i
                else:
                    pit_index = i
                    break
            pbs_current_value = pbs_event.value
        if pit_index < len(pit) - 1:
            for i in range(pit_index, len(pit)):
                pit_event = pit[i]
                pit_multiplied_by_pbs[pit_event.pos + part.start_pos] = (
                    pit_event.value * pbs_current_value
                )
        pitch_raw_data_by_part.append(pit_multiplied_by_pbs)
    pitch_raw_data: list[tuple[int, int]] = []
    for element in pitch_raw_data_by_part:
        first_pos = next(iter(element), None)
        if first_pos is None:
            continue
        first_invalid_index_in_previous = next(
            (i for i, x in enumerate(pitch_raw_data) if x[0] >= first_pos),
            None,
        )
        if first_invalid_index_in_previous is None:
            pitch_raw_data += list(element.items())
        else:
            pitch_raw_data = pitch_raw_data[:first_invalid_index_in_previous] + list(
                element.items()
            )
    data = []
    prev_pos = None
    value = None
    for pos, pitch in pitch_raw_data:
        pos_secs = synchronizer.get_actual_secs_from_ticks(pos)
        if prev_pos is not None and value is not None:
            for interp_pos in range(prev_pos + 5, pos, 5):
                interp_pos_secs = synchronizer.get_actual_secs_from_ticks(interp_pos)
                if interp_value_diff := vibrato_rate_interval_dict.get(interp_pos_secs, 0):
                    interp_value_diff *= vibrato_depth_interval_dict.get(interp_pos_secs, 64)
                data.append(
                    Point(
                        x=interp_pos,
                        y=value + round(interp_value_diff),
                    )
                )
        value = round((pitch / (PITCH_MAX_VALUE if pitch > 0 else (PITCH_MAX_VALUE + 1))) * 100)
        if value_diff := vibrato_rate_interval_dict.get(pos_secs, 0):
            value_diff *= vibrato_depth_interval_dict.get(pos_secs, 64)
        data.append(
            Point(
                x=pos,
                y=value + round(value_diff),
            )
        )
        prev_pos = pos
    pitch_simulator = PitchSimulator(
        synchronizer=synchronizer,
        portamento=PortamentoPitch.vocaloid_portamento(),
        note_list=note_list,
    )
    return (
        RelativePitchCurve(first_bar_length, lower_bound, upper_bound).to_absolute(
            data, pitch_simulator
        )
        if data and note_list
        else None
    )


def generate_for_vocaloid(
    pitch: ParamCurve,
    notes: list[Note],
    first_bar_length: int,
    synchronizer: TimeSynchronizer,
) -> VocaloidPartPitchData | None:
    pitch_simulator = PitchSimulator(
        synchronizer=synchronizer,
        portamento=PortamentoPitch.vocaloid_portamento(),
        note_list=notes,
    )
    data = RelativePitchCurve(first_bar_length).from_absolute(pitch.points.root, pitch_simulator)
    if not len(data):
        return None
    pitch_sectioned: list[list[Point]] = [[]]
    current_pos = 0
    for pitch_event in data:
        if (
            not pitch_sectioned[-1]
            or pitch_event.x - current_pos < MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS
        ):
            pitch_sectioned[-1].append(pitch_event)
        else:
            pitch_sectioned.append([pitch_event])
        current_pos = pitch_event.x
    pit: list[ControllerEvent] = []
    pbs: list[ControllerEvent] = []
    for section in pitch_sectioned:
        if len(section):
            max_abs_value = max(abs(point.y / 100) for point in section)
            pbs_for_this_section = min(math.ceil(max_abs_value), MAX_PITCH_BEND_SENSITIVITY)
            if pbs_for_this_section > DEFAULT_PITCH_BEND_SENSITIVITY:
                pbs.extend(
                    (
                        ControllerEvent(section[0][0], pbs_for_this_section),
                        ControllerEvent(
                            section[-1][0] + MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS // 2,
                            DEFAULT_PITCH_BEND_SENSITIVITY,
                        ),
                    )
                )
            else:
                pbs_for_this_section = DEFAULT_PITCH_BEND_SENSITIVITY
            pit.extend(
                ControllerEvent(
                    pitch_pos,
                    int(
                        clamp(
                            pitch_value
                            * (PITCH_MAX_VALUE if pitch_value > 0 else (PITCH_MAX_VALUE + 1))
                            / 100
                            / pbs_for_this_section,
                            -PITCH_MAX_VALUE - 1,
                            PITCH_MAX_VALUE,
                        )
                    ),
                )
                for pitch_pos, pitch_value in section
            )
    return VocaloidPartPitchData(start_pos=0, pit=pit, pbs=pbs)



import dataclasses
import re
from collections.abc import Callable

from libresvip.core.constants import DEFAULT_CHINESE_LYRIC
from libresvip.core.lyric_phoneme.chinese import CHINESE_RE
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Phones,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
    VibratoParam,
)
from libresvip.utils.translation import gettext_lazy as _

from .models import opensvip_singers, svip_note_head_tags, svip_reverb_presets
from .msnrbf.constants import (
    VALUE_LIST_VERSION_SONG_BEAT,
    VALUE_LIST_VERSION_SONG_ITRACK,
    VALUE_LIST_VERSION_SONG_NOTE,
    VALUE_LIST_VERSION_SONG_TEMPO,
)
from .msnrbf.xstudio_models import (
    MAX_NOTE_DURATION,
    MIN_NOTE_DURATION,
    XSAppModel,
    XSBeatSize,
    XSInstrumentTrack,
    XSITrack,
    XSLineParam,
    XSLineParamNode,
    XSNote,
    XSNoteHeadTag,
    XSNoteHeadTagEnum,
    XSNotePhoneInfo,
    XSReverbPreset,
    XSReverbPresetEnum,
    XSSingingTrack,
    XSSongBeat,
    XSSongTempo,
    XSVibratoPercentInfo,
    XSVibratoStyle,
)
from .options import OutputOptions


@dataclasses.dataclass
class BinarySvipGenerator:
    options: OutputOptions
    is_absolute_time_mode: bool = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_tick: int = dataclasses.field(init=False)
    first_bar_tempo: list[SongTempo] = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> tuple[str, XSAppModel]:
        version = (
            project.version
            if re.match(r"^SVIP\d\.\d\.\d$", project.version) is not None
            else "SVIP6.0.0"
        )
        model = XSAppModel()
        self.first_bar_tick = round(project.time_signature_list[0].bar_length())
        self.first_bar_tempo = [
            tempo for tempo in project.song_tempo_list if tempo.position < self.first_bar_tick
        ]
        self.is_absolute_time_mode = any(
            (tempo.bpm < 20 or tempo.bpm > 300) for tempo in project.song_tempo_list
        )
        self.synchronizer = TimeSynchronizer(
            project.song_tempo_list,
            self.first_bar_tick,
            self.is_absolute_time_mode,
            self.options.tempo,
        )

        # beat
        beat_list = model.beat_list.buf.items
        if self.is_absolute_time_mode or any(
            (beat.numerator > 255 or beat.denominator > 32) for beat in project.time_signature_list
        ):
            beat_list.append(XSSongBeat(bar_index=0, beat_size=XSBeatSize(x=4, y=4)))
        else:
            for beat in project.time_signature_list:
                beat_list.append(self.generate_time_signature(beat))
        model.beat_list.buf.size = len(beat_list)
        model.beat_list.buf.version = VALUE_LIST_VERSION_SONG_BEAT
        model.beat_list.buf_1 = model.beat_list.buf

        # tempo
        tempo_list = model.tempo_list.buf.items
        if self.is_absolute_time_mode:
            tempo_list.append(XSSongTempo(pos=0, tempo=self.options.tempo * 100))
        else:
            for tempo in project.song_tempo_list:
                tempo_list.append(self.generate_song_tempo(tempo))
        model.tempo_list.buf.size = len(tempo_list)
        model.tempo_list.buf.version = VALUE_LIST_VERSION_SONG_TEMPO
        model.tempo_list.buf_1 = model.tempo_list.buf

        # tracks
        track_list = []
        for track in project.track_list:
            ele = self.generate_track(track)
            if ele is not None:
                track_list.append(ele)
        model.track_list.size = len(track_list)
        model.track_list.version = VALUE_LIST_VERSION_SONG_ITRACK
        model.track_list.items = track_list
        return version, model

    @staticmethod
    def generate_song_tempo(tempo: SongTempo) -> XSSongTempo:
        return XSSongTempo(pos=tempo.position, tempo=round(tempo.bpm * 100))

    @staticmethod
    def generate_time_signature(signature: TimeSignature) -> XSSongBeat:
        return XSSongBeat(
            bar_index=signature.bar_index,
            beat_size=XSBeatSize(x=signature.numerator, y=signature.denominator),
        )

    def generate_track(self, track: Track) -> XSITrack | None:
        s_track: XSITrack | None = None
        if isinstance(track, SingingTrack):
            singer_id = opensvip_singers.get_id(track.ai_singer_name)
            if singer_id == "":
                singer_id = opensvip_singers.get_id(self.options.singer)
            s_track = XSSingingTrack(
                ai_singer_id=singer_id,
                reverb_preset=XSReverbPreset(
                    svip_reverb_presets.get(track.reverb_preset, XSReverbPresetEnum.NONE)
                ),
            )

            note_list = s_track.note_list.buf.items
            for note in track.note_list:
                if (new_note := self.generate_note(note)) is not None:
                    note_list.append(new_note)
            s_track.note_list.buf.size = len(note_list)
            s_track.note_list.buf.version = VALUE_LIST_VERSION_SONG_NOTE
            s_track.note_list.buf_1 = s_track.note_list.buf

            params = self.generate_params(track.edited_params)
            s_track.edited_pitch_line = params["Pitch"]
            s_track.edited_volume_line = params["Volume"]
            s_track.edited_breath_line = params["Breath"]
            s_track.edited_gender_line = params["Gender"]
            if self.options.version == "SVIP7.0.0":
                s_track.edited_power_line = params["Strength"]
        elif isinstance(track, InstrumentalTrack):
            s_track = XSInstrumentTrack(
                instrument_file_path=track.audio_file_path,
                offset_in_pos=track.offset,
            )
        else:
            return None
        s_track.name = track.title
        s_track.mute = track.mute
        s_track.solo = track.solo
        s_track.volume = track.volume
        s_track.pan = track.pan
        return s_track

    def generate_note(self, note: Note) -> XSNote | None:
        if not note.lyric and not note.pronunciation:
            return None
        xs_note = XSNote(
            start_pos=round(self.synchronizer.get_actual_ticks_from_ticks(note.start_pos)),
            key_index=note.key_number + 12,
            head_tag=XSNoteHeadTag(
                value=svip_note_head_tags.get(note.head_tag, XSNoteHeadTagEnum.NoTag)
            ),
            lyric=note.lyric if CHINESE_RE.match(note.lyric) is not None else DEFAULT_CHINESE_LYRIC,
            pronouncing=note.pronunciation or "",
        )
        xs_note.width_pos = (
            round(self.synchronizer.get_actual_ticks_from_ticks(note.end_pos)) - xs_note.start_pos
        )
        note_duration = self.synchronizer.get_duration_secs_from_ticks(note.start_pos, note.end_pos)
        if note_duration < MIN_NOTE_DURATION:
            msg_prefix = _("Note duration is too short:")
            show_warning(f"{msg_prefix} {note.lyric}")
        elif note_duration > MAX_NOTE_DURATION:
            msg_prefix = _("Note duration is too long:")
            show_warning(f"{msg_prefix} {note.lyric}")
        if note.edited_phones is not None:
            xs_note.note_phone_info = self.generate_phones(note.edited_phones)
        if note.vibrato is not None:
            percent, vibrato = self.generate_vibrato(note.vibrato)
            xs_note.vibrato_percent_info = percent
            xs_note.vibrato = vibrato
        return xs_note

    @staticmethod
    def generate_phones(edited_phones: Phones) -> XSNotePhoneInfo:
        return XSNotePhoneInfo(
            head_phone_time_in_sec=edited_phones.head_length_in_secs,
            mid_part_over_tail_part_ratio=edited_phones.mid_ratio_over_tail,
        )

    def generate_vibrato(
        self, vibrato: VibratoParam
    ) -> tuple[XSVibratoPercentInfo, XSVibratoStyle]:
        percent = XSVibratoPercentInfo(
            start_percent=vibrato.start_percent,
            end_percent=vibrato.end_percent,
        )
        vibrato_style = XSVibratoStyle(
            is_anti_phase=vibrato.is_anti_phase,
            amp_line=self.generate_param_curve(
                vibrato.amplitude, left=-1, right=100001, is_ticks=False
            ),
            freq_line=self.generate_param_curve(
                vibrato.frequency, left=-1, right=100001, is_ticks=False
            ),
        )
        return percent, vibrato_style

    def generate_params(self, edited_params: Params) -> dict[str, XSLineParam]:
        return {
            "Pitch": self.generate_param_curve(
                edited_params.pitch,
                op=lambda x: x + 1150 if x > -100 else -100,
            ),
            "Volume": self.generate_param_curve(edited_params.volume),
            "Breath": self.generate_param_curve(edited_params.breath),
            "Gender": self.generate_param_curve(edited_params.gender),
            "Strength": self.generate_param_curve(edited_params.strength),
        }

    def generate_param_curve(
        self,
        param_curve: ParamCurve,
        op: Callable[[float], float] | None = None,
        left: int = -192000,
        right: int = 1073741823,
        termination: int = 0,
        is_ticks: bool = True,
    ) -> XSLineParam:
        if op is None:

            def _op(x: float) -> float:
                return x

            op = _op

        line = XSLineParam()
        # param_curve.points = sorted(param_curve.points, key=operator.attrgetter("x"))
        for p in param_curve.points.root:
            if left <= p.x <= right:
                if self.is_absolute_time_mode and is_ticks and p.x != left and p.x != right:
                    pos = (
                        round(
                            self.synchronizer.get_actual_ticks_from_ticks(p.x - self.first_bar_tick)
                        )
                        + self.first_bar_tick
                    )
                else:
                    pos = p.x
                node = XSLineParamNode(pos=pos, value=int(op(p.y)))
                line.nodes.append(node)
        if len(line.nodes) == 0 or line.nodes[0].pos > left:
            bound = XSLineParamNode(pos=left, value=termination)
            line.nodes.insert(0, bound)
        if len(line.nodes) == 0 or line.nodes[-1].pos < right:
            bound = XSLineParamNode(pos=right, value=termination)
            line.nodes.append(bound)
        line.convert_to_param()
        return line
